"""
Software for controlling the simulated DQC hardware.

Designed to act on hardware created using
:mod: `~dqc_simulator.hardware.dqc_creation`
"""

import pydynaa
from netsquid.components import instructions as instr
from netsquid.components.qprogram import QuantumProgram
from netsquid.protocols.nodeprotocols import NodeProtocol, LocalProtocol

from dqc_simulator.hardware.quantum_processors import QPU
from dqc_simulator.software.compilers import (
    sort_greedily_by_node_and_time,
    sort_many_qpus_greedily_by_node_and_time,
)
from dqc_simulator.software.physical_layer import (
    PhysicalLayerProtocol,
    AbstractCentralSourceEntangleProtocol,
)
from dqc_simulator.util.helper import QDCSignals


class UnfinishedQuantumCircuitError(Exception):
    """
    Error to be raised when a quantum circuit has not run all the way through.
    """

    pass


class InterpreterProtocol(NodeProtocol):
    """
    Interprets and runs all operations for a given QPU.

    This protocol will be used as a subprotocol of another protocol
    (the `superprotocol`, DQCMasterProtocol).

    Parameters
    ----------
    superprotocol : :class:`~netsquid.protocols.protocol.Protocol`
        The protocol that this will be a subprotocol of.
    gate_tuples : list of lists of tuples
        The gates and primitives (blocks of gates and operations that are part
        of remote gates or other DQC-specific operations) to be conducted. Each
        sublist corresponds to the gates to be
        conducted on a time slice. The tuples represent the gates or primitives.
        Remote gates may have a list of
        tuples as the first entry of a gate tuple if in the recv role to
        allow several local
        gates to be applied as part of the remote gate.
        Each gate_tuple in gate_tuples4this_node_and_time must have
        form:
        1) for single-qubit gate: (gate_type, qubit_index)
        2) for single-qubit primitive: (gate_Type, qubit_index, primitive_label)
        3) for two-qubit gate: (gate_type, qubit_index1, qubit_index2)
        4) for remote gate primitive:
                either : i) (data_qubit_index or sometimes
                             index of qubit to be
                             teleported/distributed,
                             other_node_name, scheme, role).
                                     OR
                         ii) (other_node_name, scheme, role)


            where othernode_name is the name attribute of a
            netsquid.nodes.Node object, and gate_type is an
            instruction from netsquid.components.instructions
    node : :class:`netsquid.nodes.node.Node`, subclass thereof or None, optional
        The QPU node that this protocol will act on. If None, a node should be
        set later before starting this protocol.
    name : str or None, optional
        Name of protocol. If None, the name of the class is used.

    Notes
    -----
    The logic used in the _run_time_slice and, by extension, run methods of
    this protocol is a generalisation of the use of subprotocols in NetSquid
    `~netsquid.protocols.protocol.Protocol`s. Subprotocols are
    `~netsquid.protocols.protocol.Protocol`s and so their run method can only
    yield `~pydynaa.core.EventExpression`s. This is clean because it is easy to
    understand what is being yielded. However, here it was convenient to
    make use of the fact that `Protocol` run methods are merely a specific
    type of the more general python generator object. Generators can yield
    anything and so rather than using `Protocol`s, whose run methods can yield
    only `EventExpressions`, I instead use generators that yield
    `EventExpression`s, to indicate that something has happened in the sim,
    AND variables that can be used by subsequent generators.

    If subprotocols had been used it would have been necessary to send
    variables, such as recently allocated comm-qubit indices and the
    `QuantumProgram` being modified, as signals between the different
    subprotocols. It becomes very complicated to make sure that subprotocols
    which implement later instructions are actually started before the
    subprotocols implementing earlier instructions so that the latter
    subprotocols can wait on signals from the former. Ie, things like the
    last used comm-qubit index become very hard to track. This is particularly
    challenging to implement inside a simple if elif control flow. By contrast,
    with subgenerators, the relevant variables are returned after all
    `EventExpression`s have been yielded, making the logic simpler.
    """

    # The core idea of what follows is to add instructions to a QuantumProgram
    # until remote gates occur, at which point the program will be run to that
    # point, re-initialised and then added to with instructions needed for the
    # part of the remote gate done on this node. The classical port names have
    # been chosen so that the sending and receiving port have the same f-string
    # but will have different names when this protocol is acted on the
    # appropriate nodes

    def __init__(self, superprotocol, gate_tuples, node=None, name=None):
        super().__init__(node, name)
        # the allowed subgenerators (similar to subprotocols, see Notes section
        # above):
        self.single_qubit_subroutines = {
            "logged": self._logged_instr,
        }
        self.remote_gate_variants = {
            "cat": self._cat_subgenerator,
            "tp": self._tp_subgenerator,
        }
        # other import attributes
        self.superprotocol = superprotocol
        self.gate_tuples = gate_tuples
        # adding attributes for tracking what gates have evaluated
        self._gate_tuples_evaluated = [[]]
        # the next attribute is for storing gate tuples for local gates added to
        # a QuantumProgram but not yet evaluated
        self._local_gate_tuples_pending = []
        self._max_num_time_slices = len(self.gate_tuples)
        self.ent_request_label = "ENT_REQUEST"
        self.ent_ready_label = "ENT_READY"
        self.ent_failed_label = "ENT_FAILED"
        self.start_time_slice_label = "START_TIME_SLICE"
        self.finished_time_slice_label = "FINISHED_TIME_SLICE"
        self.add_signal(self.ent_request_label)
        self.add_signal(self.ent_ready_label)
        self.add_signal(self.ent_failed_label)
        self.add_signal(self.start_time_slice_label)
        self.add_signal(self.finished_time_slice_label)
        self.add_signal(QDCSignals.RESULT_PRODUCED)

    def add_phys_layer(self, physical_layer_protocol):
        """
        Convenience method for adding physical layer.

        Parameters
        ----------
        physical_layer_protocol : :class:`~dqc_simulator.software.physical_layer.PhysicalLayerProtocol`
            A physical layer protocol.
        """
        physical_layer_protocol.superprotocol = self
        self.add_subprotocol(physical_layer_protocol, name="physical_layer_protocol")

    def _flexi_program_apply(
        self, qprogram, gate_instr, qubit_indices, gate_op, output_key="last"
    ):
        """
        Handles the possibility gate_op is not defined in the gate instruction.

        Such a scenario can be useful because it
        allows one to define one PhysicalInstruction for a variety of gates,
        making it possible to implement gates of the form
        gate_name(list_of_parameters) without having to define a
        PhysicalInstruction for infinitely many possible parameter values.

        Parameters
        ----------
        qprogram : an instance of
                   netsquid.netsquid.components.qprogram.QuantumProgram
            A quantum program.
        gate_instr : instance of netsquid.components.instructions.Instruction
            A quantum instruction describing a gate. Note, this may not
            have the operator used to carry out the gate predefined. If the
            operator is not incorporated in the instruction definition it is
            included in gate_op (see below).
        qubit_indices : int or list of int
            The indices of the qubits to act the instruction on.
        gate_op : instance of netsquid.qubits.operators.Operator
            The operator used to apply the quantum gate.
        output_key : str, optional
            Dictionary key to store instruction output with. Default
            is to store all output using the key last.

        .. todo::

            Handle situations where entanglement is successful or not
            successful in _request_entanglement method (see comment there)
        """
        if gate_op is None:
            qprogram.apply(gate_instr, qubit_indices, output_key=output_key)
        else:
            qprogram.apply(
                gate_instr, qubit_indices, operator=gate_op, output_key=output_key
            )

    def _request_entanglement(
        self,
        role,
        other_node_name,
        comm_qubit_indices,
        num_entanglements2generate=1,
        entanglement_type2generate="bell_pair",
    ):
        """
        Requests entanglement.

        Passes entanglement request to link layer.

        Parameters
        ----------
        role : str
            Whether the QPU is a 'client' (initiating the handshake) or a
            'server' (responding to the handshake.)
        other_node_name : str
            The name of the other QPU node between which entanglement should
            be distributed.
        comm_qubit_indices : list of int
            The indices of the comm-qubits that should be used.
        num_entanglements2generate : int, optional
            The number of entanglements that should be distributed. Eg, for
            bipartite entanglement this would be the number of entangled pairs.
            The default is 1.
        entanglement_type2generate : str, optional
            The type of entanglement to distribute. The default is 'bell_pair'.

        Yields
        ------
        None.

        """
        # send entanglement request to link layer specifying the other node
        # involved in the connection and the comm-qubit to be used:
        self.send_signal(
            self.ent_request_label,
            result=(
                role,
                other_node_name,
                comm_qubit_indices,
                num_entanglements2generate,
                entanglement_type2generate,
            ),
        )
        # wait for entangled qubit to arrive in requested comm-qubit slot:
        wait4succsessful_ent = self.await_signal(
            self.subprotocols["physical_layer_protocol"],
            signal_label=self.ent_ready_label,
        )
        wait4failed_ent = self.await_signal(
            self.subprotocols["physical_layer_protocol"],
            signal_label=self.ent_failed_label,
        )
        evexpr = yield wait4succsessful_ent | wait4failed_ent
        # TO DO: handle the case where entanglement fails properly rather than
        # raising an error.
        if evexpr.second_term.value:  # if entanglement failed signal received
            raise NotImplementedError(
                "Entanglement distribution has failed. If a using a "
                "non-deterministic physical layer, this is most likely "
                "because the max_num_ent_attempts has been "
                "exceeded. At the moment, there is no handler in place "
                "to deal with this scenario."
            )

    def _cat_entangle(
        self, program, other_node_name, data_or_tele_qubit_index, classical_comm_port
    ):
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []

        program = QuantumProgram()
        if len(self.node.qmemory.comm_qubits_free) == 0:
            # if there are NO comm_qubits free:
            # using first comm-qubit that's free with a
            # name unique to this role so that it can
            # be reused in the "disentangle_end" role
            raise Exception(
                f"Scheduling error: No"
                f"comm-qubits free. There"
                f" are too many remote"
                f"gates on {self.node.name}"
                f"in this time slice"
            )
        else:
            comm_qubit_index = self.node.qmemory.comm_qubits_free[0]
            # removing comm qubit being used from list
            # of available comm-qubits:
            if not self.node.qmemory.ebit_ready:  # if there is no ebit ready:
                yield from self._request_entanglement(
                    "client", other_node_name, [comm_qubit_index]
                )
            program.apply(
                instr.INSTR_CNOT, [data_or_tele_qubit_index, comm_qubit_index]
            )
            program.apply(instr.INSTR_MEASURE, [comm_qubit_index], output_key="ma")
            # run program to this point
            yield self.node.qmemory.execute_program(program)
            self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
            # re-initialising self._local_gate_tuples_pending
            self._local_gate_tuples_pending = []
            # comm-qubit is immediately freed by
            # measurement so is not deleted from
            # comm_qubits_free here
            (ma,) = program.output["ma"]
            classical_comm_port.tx_output(ma)
            program = QuantumProgram()  # resetting
            return program  # note that this will only be
            # be returned after all the
            # EventExpressions from the prev.
            # yield statements have been
            # outputed

    def _cat_correct(self, program, other_node_name, classical_comm_port):
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []
        program = QuantumProgram()
        if len(self.node.qmemory.comm_qubits_free) == 0:
            # if there are NO comm_qubits free:
            raise Exception(
                f"Scheduling error: No"
                f" comm-qubits free. There"
                f" are too many remote"
                f" gates on"
                f" {self.node.name} in "
                f"this time slice"
            )
        else:  # if there are comm-qubits free
            # using first comm_qubit that's free with
            # name specific to role (so that it can be
            # found by "disentangle_start")
            comm_qubit_index = self.node.qmemory.comm_qubits_free[0]
            # removing comm qubit being used from list
            # of available comm-qubits:
            if not self.node.qmemory.ebit_ready:  # if there is no ebit ready:
                yield from self._request_entanglement(
                    "server", other_node_name, [comm_qubit_index]
                )
            del self.node.qmemory.comm_qubits_free[0]
            # wait for measurement result from _cat_entangle on the other node
            yield self.await_port_input(classical_comm_port)
            (meas_result,) = classical_comm_port.rx_input().items
            if meas_result == 1:
                program.apply(instr.INSTR_X, comm_qubit_index)
            elif meas_result != 0:
                raise ValueError(
                    "Measurement result must be an integer with "
                    f"value 0 or 1. The current value, "
                    f"{meas_result}, does not meet this criteria."
                )
        return (comm_qubit_index, program)  # note that this will only be
        # be returned after all the
        # EventExpressions from the prev.
        # yield statements have been
        # outputed

    def _cat_disentangle_start(self, program, comm_qubit_index, classical_comm_port):
        # don't need comm-qubit free to do this. It
        # will use the comm-qubit index defined in the
        # last operation (ie, correct)
        program.apply(instr.INSTR_H, comm_qubit_index)
        program.apply(instr.INSTR_MEASURE, comm_qubit_index, output_key="mb")
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []
        (meas_result,) = program.output["mb"]
        program = QuantumProgram()  # re-initialising
        classical_comm_port.tx_output(meas_result)
        # freeing comm-qubit now that it is not needed
        self.node.qmemory.comm_qubits_free = self.node.qmemory.comm_qubits_free + [
            comm_qubit_index
        ]
        # re-odering from smallest number to largest
        self.node.qmemory.comm_qubits_free.sort()
        return program

    def _cat_disentangle_end(
        self, program, classical_comm_port, data_or_tele_qubit_index
    ):
        yield self.await_port_input(classical_comm_port)
        (meas_result,) = classical_comm_port.rx_input().items
        if meas_result == 1:
            # correcting some qubit entangled in the
            # cat-entanglement step
            program.apply(instr.INSTR_Z, [data_or_tele_qubit_index])
            yield self.node.qmemory.execute_program(program)
            self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
            # re-initialising self._local_gate_tuples_pending
            self._local_gate_tuples_pending = []
            program = QuantumProgram()
        elif meas_result != 0:
            raise ValueError(
                "Measurement result must be an integer with value"
                f" 0 or 1. The current value, {meas_result}, does "
                "not meet this criteria."
            )
        # if meas_result == 0, we do nothing. No gates are required.
        return program

    def _bsm(
        self, program, data_or_tele_qubit_index, other_node_name, classical_comm_port
    ):
        # doing all pending local operations
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []

        # implement a BSM
        program = QuantumProgram()
        if len(self.node.qmemory.comm_qubits_free) == 0:
            # if there are no comm-qubits free:
            raise Exception(
                f"Scheduling error:"
                f"No comm-qubits free."
                f" There are too many "
                f"remote gates on "
                f" {self.node.name} in"
                f" this time slice"
            )
        else:  # if there are comm-qubits free
            comm_qubit_index = self.node.qmemory.comm_qubits_free[0]
            # there is no need to delete a comm-qubit
            # it will be restored at the end of this
            # block
            if data_or_tele_qubit_index == -1 and len(
                self.node.qmemory.comm_qubits_free
            ) < len(self.node.qmemory.comm_qubit_positions):
                # letting qubit to be teleported be the
                # last used communication qubit from the
                # prev time-slice
                data_or_tele_qubit_index = comm_qubit_index - 1
            if not self.node.qmemory.ebit_ready:
                yield from self._request_entanglement(
                    "client", other_node_name, [comm_qubit_index]
                )
            program.apply(
                instr.INSTR_CNOT, [data_or_tele_qubit_index, comm_qubit_index]
            )
            program.apply(instr.INSTR_H, [data_or_tele_qubit_index])
            program.apply(instr.INSTR_MEASURE, [comm_qubit_index], output_key="m1")
            # =============================================================================
            #                                     #discarding the comm qubit to allow
            #                                     #another one to be retrieved.
            #                                     #Physically, this can be thought of as
            #                                     #reinitialising
            #                                     #the comm-qubit prior to entangling it
            #                                     #with another photon
            #                                     program.apply(instr.INSTR_DISCARD,
            #                                                   comm_qubit_index)
            # =============================================================================
            program.apply(
                instr.INSTR_MEASURE, [data_or_tele_qubit_index], output_key="m2"
            )
            yield self.node.qmemory.execute_program(program)
            self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
            # re-initialising self._local_gate_tuples_pending
            self._local_gate_tuples_pending = []
            (m1,) = program.output["m1"]
            (m2,) = program.output["m2"]
            program = QuantumProgram()
            # re-setting data-qubit to zero after
            # measurement (comm-qubit already reset
            # when entanglement re-distributed)
            # TO DO: GET THIS TO TAKE SAME TIME AS
            # APPLYING AN X-GATE controlled by the
            # measurement result.
            program.apply(instr.INSTR_INIT, data_or_tele_qubit_index)
            yield self.node.qmemory.execute_program(program)
            self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
            # re-initialising self._local_gate_tuples_pending
            self._local_gate_tuples_pending = []
            program = QuantumProgram()
            classical_comm_port.tx_output((m1, m2))
            if data_or_tele_qubit_index in self.node.qmemory.comm_qubit_positions:
                # freeing comm_qubit now that it has been
                # measured
                self.node.qmemory.comm_qubits_free = (
                    self.node.qmemory.comm_qubits_free + [data_or_tele_qubit_index]
                )
                # re-ordering from smallest number to
                # largest
                self.node.qmemory.comm_qubits_free.sort()
        return program

    def _handle_meas_results4tp_correct(self, meas_results, index, program):
        error_msg = (
            "All elements of meas_results must be integers with value"
            " 0 or 1. The current meas_results, which have the value,"
            f" {meas_results} do not meet this criteria"
        )
        if meas_results[0] == 1:
            program.apply(instr.INSTR_X, index)
        elif meas_results[0] != 0:
            raise ValueError(error_msg)
        # elif meas_results[0] == 0, do nothing
        if meas_results[1] == 1:
            program.apply(instr.INSTR_Z, index)
        elif meas_results[1] != 0:
            raise ValueError(error_msg)
        # elif meas_results[1] == 0, do nothing

    def _tp_correct(
        self,
        program,
        other_node_name,
        classical_comm_port,
        reserve_comm_qubit,
        swap_commNdata,
        data_or_tele_qubit_index=None,
        manual=False,
    ):
        """


        Parameters
        ----------
        program : TYPE
            DESCRIPTION.
        other_node_name : TYPE
            DESCRIPTION.
        classical_comm_port : TYPE
            DESCRIPTION.
        reserve_comm_qubit : bool
            Whether or not to reserve the comm-qubit being used, so that the
            state teleported there cannot be overwritten. Will reserve if True.
        swap_commNdata : bool
            Whether to swap the data qubit and comm qubit prior to applying
            measurement dependent gates.
        manual : bool
            Whether to have the user manually specify the comm-qubits or do
            this automatically (False). The default is False.
        data_or_tele_qubit_index : int or None
            The index of the data qubit index to SWAP the comm qubit index with
            if swap_commNdata is True or the index of the qubit that has been
            teleported to if manual is True. The default is None.

        Raises
        ------
        Exception
            DESCRIPTION.

        Yields
        ------
        TYPE
            DESCRIPTION.
        TYPE
            DESCRIPTION.

        """
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []
        program = QuantumProgram()
        if len(self.node.qmemory.comm_qubits_free) == 0:
            # if there are no comm-qubits free:
            raise Exception(
                f"Scheduling error: No "
                f" comm-qubits free. "
                f"There are too many"
                f" remote gates on "
                f"{self.node.name} in "
                f"this time slice"
            )
        # elif there are comm-qubits free:

        # establishing new comm-qubit for the current teleportation
        if not manual:  # if specifying comm-qubit index automatically:
            comm_qubit_index = self.node.qmemory.comm_qubits_free[0]
        else:  # if specifying comm-qubit index manually:
            comm_qubit_index = data_or_tele_qubit_index

        if reserve_comm_qubit:
            # removing comm qubit being used from list of available comm-qubits
            reserved_index = self.node.qmemory.comm_qubits_free.index(comm_qubit_index)
            del self.node.qmemory.comm_qubits_free[reserved_index]
        if not self.node.qmemory.ebit_ready:  # if there is no ebit ready:
            yield from self._request_entanglement(
                "server", other_node_name, [comm_qubit_index]
            )
        yield self.await_port_input(classical_comm_port)
        (meas_results,) = classical_comm_port.rx_input().items
        if swap_commNdata:
            # =============================================================================
            #                 program.apply(instr.INSTR_CNOT, [comm_qubit_index,
            #                                                  data_or_tele_qubit_index])
            #                 program.apply(instr.INSTR_CNOT, [data_or_tele_qubit_index,
            #                                                  comm_qubit_index])
            #                 program.apply(instr.INSTR_CNOT, [comm_qubit_index,
            #                                                  data_or_tele_qubit_index])
            # =============================================================================
            program.apply(
                instr.INSTR_SWAP, [comm_qubit_index, data_or_tele_qubit_index]
            )
            self._handle_meas_results4tp_correct(
                meas_results, data_or_tele_qubit_index, program
            )
        else:
            self._handle_meas_results4tp_correct(
                meas_results, comm_qubit_index, program
            )
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []
        program = QuantumProgram()
        return (comm_qubit_index, program)

    def _cat_subgenerator(
        self,
        role,
        program,
        other_node_name,
        data_or_tele_qubit_index,
        classical_comm_port,
        comm_qubit_index,
    ):
        # =============================================================================
        #         #commented out block is alternative implementation. It is more easily
        #         #extensible but I think maybe a little harder to understand, in the
        #         #sense that it hides which subgenerators change what variables:
        #         cat_subgenerator_funcs = {
        #                              "entangle" : self._cat_entangle(other_node_name,
        #                                                              comm_qubit_index,
        #                                                              ent_request_port),
        #                              "correct" : self._cat_correct(program,
        #                                                            other_node_name,
        #                                                            classical_comm_port,
        #                                                            ent_request_port),
        #                              "disentangle_start" : self._cat_disentangle_start(
        #                                                         program,
        #                                                         comm_qubit_index,
        #                                                         classical_comm_port),
        #                              "disentangle_end" : self._cat_disentangle_end(
        #                                                     program,
        #                                                     classical_comm_port,
        #                                                     data_or_tele_qubit_index)}
        #         evexpr_or_vars = yield from self.cat_subgenerator_funcs[role]
        # =============================================================================

        if role == "entangle":
            evexpr_or_program = yield from self._cat_entangle(
                program, other_node_name, data_or_tele_qubit_index, classical_comm_port
            )
            # output will be EventExpression until end of
            # generator is reached when it will instead
            # be tuple of useful variables to pass on to
            # subsequent code. The following works on the
            # tuple:
            program = evexpr_or_program

        elif role == "correct":
            evexpr_or_tuple = yield from self._cat_correct(
                program, other_node_name, classical_comm_port
            )
            # output will be EventExpression until end of
            # generator is reached when it will instead
            # be tuple of useful variables to pass on to
            # subsequent code. The following works on the
            # tuple:
            comm_qubit_index = evexpr_or_tuple[0]
            program = evexpr_or_tuple[1]

        elif role == "disentangle_start":
            evexpr_or_program = yield from self._cat_disentangle_start(
                program, comm_qubit_index, classical_comm_port
            )
            program = evexpr_or_program

        elif role == "disentangle_end":
            evexpr_or_program = yield from self._cat_disentangle_end(
                program, classical_comm_port, data_or_tele_qubit_index
            )
            program = evexpr_or_program

        else:
            raise ValueError("final element of gate_tuple is not valid role")
        return (comm_qubit_index, program)

    def _tp_subgenerator(
        self,
        role,
        program,
        other_node_name,
        data_or_tele_qubit_index,
        classical_comm_port,
        comm_qubit_index,
    ):
        if role == "bsm":
            evexpr_or_program = yield from self._bsm(
                program, data_or_tele_qubit_index, other_node_name, classical_comm_port
            )
            program = evexpr_or_program

        elif role == "correct":
            evexpr_or_tuple = yield from self._tp_correct(
                program,
                other_node_name,
                classical_comm_port,
                reserve_comm_qubit=True,
                swap_commNdata=False,
            )
            comm_qubit_index = evexpr_or_tuple[0]
            program = evexpr_or_tuple[1]

        elif role == "correct4tele_only":
            evexpr_or_tuple = yield from self._tp_correct(
                program,
                other_node_name,
                classical_comm_port,
                reserve_comm_qubit=False,
                swap_commNdata=False,
            )
            comm_qubit_index = evexpr_or_tuple[0]
            program = evexpr_or_tuple[1]

        elif role == "swap_then_correct":
            evexpr_or_tuple = yield from self._tp_correct(
                program,
                other_node_name,
                classical_comm_port,
                reserve_comm_qubit=False,
                swap_commNdata=True,
                data_or_tele_qubit_index=data_or_tele_qubit_index,
            )
            comm_qubit_index = evexpr_or_tuple[0]
            program = evexpr_or_tuple[1]

        elif role == "correct_manually":
            evexpr_or_tuple = yield from self._tp_correct(
                program,
                other_node_name,
                classical_comm_port,
                reserve_comm_qubit=True,
                swap_commNdata=False,
                manual=True,
                data_or_tele_qubit_index=data_or_tele_qubit_index,
            )
            comm_qubit_index = evexpr_or_tuple[0]
            program = evexpr_or_tuple[1]

        else:
            raise ValueError(f"final element, {role} of gate_tuple is not a valid role")
        return (comm_qubit_index, program)

    # SHOULD the following be implemented as a protocol in its own right along
    # with all primitives?
    def _logged_instr(self, program, gate_tuple):
        """
        Subgenerator that carries out a single-qubit subroutine in which a
        single-qubit gate is conducted and the result logged.

        This is most likely to be used for a measurements of some type (ie, the
        single-qubit gate is a measurement in some basis).

        Parameters
        ----------
        gate_instr : `~netsquid.components.instructions.Instruction`
            The single-qubit gate to carry out a logged version of. In most
            cases this will be measurement in some basis.
        qubit_index : int
            The qubit to act on.

        Yields
        -------
        `~pydynaa.core.EventExpression`
            These allow the sim to wait on Events to happen.

        """
        gate_instr, gate_op = self._get_gate_instruction_and_op(gate_tuple)
        ancilla_qubit_index = gate_tuple[1]
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []
        program = QuantumProgram()
        self._flexi_program_apply(
            program, gate_instr, ancilla_qubit_index, gate_op, output_key="result"
        )
        yield self.node.qmemory.execute_program(program)
        (result,) = program.output["result"]
        self.send_signal(
            QDCSignals.RESULT_PRODUCED, result=(result, ancilla_qubit_index)
        )
        self._gate_tuples_evaluated[-1].append(gate_tuple)

    def _get_gate_instruction_and_op(self, gate_tuple):
        gate_instr = gate_tuple[0]
        # handling gate types with associated parameters:
        try:  # if gate_instr is iterable:
            gate_op = gate_instr[1]
            gate_instr = gate_instr[0]
        except TypeError:  # elif gate_instr is not iterable
            gate_op = None
        return gate_instr, gate_op

    def _distribute_ebit(self, program, gate_tuple):
        """
        Low-level command to distribute an ebit (pair of entangled qubits)

        Parameters
        ----------
        program : TYPE
            DESCRIPTION.
        gate_tuple : TYPE
            DESCRIPTION.

        Yields
        ------
        TYPE
            DESCRIPTION.

        """
        comm_qubit_index = self.node.qmemory.comm_qubits_free[0]
        other_node_name = gate_tuple[0]
        role = gate_tuple[1]
        yield from self._request_entanglement(
            role,
            other_node_name,
            [comm_qubit_index],
            num_entanglements2generate=1,
            entanglement_type2generate="bell_pair",
        )
        # Reserving comm-qubit
        del self.node.qmemory.comm_qubits_free[0]
        self._gate_tuples_evaluated[-1].append(gate_tuple)

    def _handle_single_qubit_gate(self, program, gate_tuple):
        gate_instr, gate_op = self._get_gate_instruction_and_op(gate_tuple)
        qubit_index = gate_tuple[1]
        self._flexi_program_apply(program, gate_instr, qubit_index, gate_op)
        self._local_gate_tuples_pending.append(gate_tuple)
        # self._local_gate_tuples_pending will be added to
        # self._gate_tuples_evaluated whenever a program is run,
        # allowing us to raise error if not all gate tuples are
        # evaluated at the end of the sim run.

    def _handle_remote_gate_primitive(self, program, gate_tuple, comm_qubit_index):
        # The remote gates in this block will use different
        # comm-qubits because logical gates using the same
        # comm-qubit will occur in different time slices or will
        # have been converted to local gates once teleportation
        # or cat-entanglement has already been done

        if len(gate_tuple) == 4:
            data_or_tele_qubit_index = gate_tuple[0]
            other_node_name = gate_tuple[1]
            scheme = gate_tuple[2]
            role = gate_tuple[3]
        elif len(gate_tuple) == 3:
            data_or_tele_qubit_index = None
            other_node_name = gate_tuple[0]
            scheme = gate_tuple[1]
            role = gate_tuple[2]
        node_names = [self.node.name, other_node_name]
        node_names.sort()
        classical_connection_port_name = self.node.connection_port_name(
            other_node_name, label="classical"
        )
        classical_conn_port = self.node.ports[classical_connection_port_name]
        evexpr_or_variables = yield from self.remote_gate_variants[scheme](
            role,
            program,
            other_node_name,
            data_or_tele_qubit_index,
            classical_conn_port,
            comm_qubit_index,
        )
        # past this point evexpr_or_variables will be the
        # variables
        variables4subsequent_local_gates = evexpr_or_variables
        # keeping track of what gate tuples have been evaluated
        # to raise error if not all evaluated at end of sim
        self._gate_tuples_evaluated[-1].append(gate_tuple)
        return variables4subsequent_local_gates

    def _handle_local_two_qubit_gate(self, program, gate_tuple, comm_qubit_index):
        gate_instr, gate_op = self._get_gate_instruction_and_op(gate_tuple)
        qubit_index0 = gate_tuple[1]
        qubit_index1 = gate_tuple[2]
        if qubit_index0 == -1:
            qubit_index0 = comm_qubit_index  # defined in remote gate
            # section above
        self._flexi_program_apply(
            program, gate_instr, [qubit_index0, qubit_index1], gate_op
        )
        self._local_gate_tuples_pending.append(gate_tuple)
        # self._local_gate_tuples_pending will be added to
        # self._gate_tuples_evaluated whenever a program is run,
        # allowing an error to be raised if not all gate tuples are
        # evaluated at the end of the sim run.

    def _run_time_slice(self, gate_tuples4time_slice):
        # intialising variables that should be overwritten later but are needed
        # as function arguments:
        program = QuantumProgram()
        comm_qubit_index = float("nan")
        for gate_tuple in gate_tuples4time_slice:
            if len(gate_tuple) == 2:  # if single-qubit gate:
                self._handle_single_qubit_gate(program, gate_tuple)
            elif gate_tuple[-1] in self.single_qubit_subroutines:
                yield from self.single_qubit_subroutines[gate_tuple[-1]](
                    program, gate_tuple
                )
            elif gate_tuple[-1] == "distribute_ebit":
                yield from self._distribute_ebit(program, gate_tuple)
            elif (
                isinstance(gate_tuple[-1], str)
                and gate_tuple[-2] in self.remote_gate_variants
            ):  # if primitive for remote
                # gate:
                # The remote gates in this block will use different
                # comm-qubits because logical gates using the same
                # comm-qubit will occur in different time slices or will
                # have been converted to local gates once teleportation
                # or cat-entanglement has already been done
                evexpr_or_variables = yield from self._handle_remote_gate_primitive(
                    program, gate_tuple, comm_qubit_index
                )
                # past this point evexpr_or_variables will be the
                # variables for use in subsequent remote gates.
                comm_qubit_index = evexpr_or_variables[0]
                # The program needs explicitly returned here and not in
                # other subgenerators because of the additional nesting of
                # self._flexi_program_apply inside _cat_subgenerators or
                # _tp_subgenerators which are in turn inside of
                # _handle_remote_gate_primitive.
                program = evexpr_or_variables[1]
            # strictly speaking the following will allow any local
            # multi-qubit gate. You need to decide whether you want to allow
            # for this or not (if you want arbitrary multi-qubit then need
            # to change compiler)
            elif isinstance(gate_tuple[-1], int):  # if local 2-qubit gate
                self._handle_local_two_qubit_gate(program, gate_tuple, comm_qubit_index)
        # executing any instructions that have not yet been executed. It is
        # important that the quantum program is always reset after each node
        # is finished to avoid waiting infinitely long if there is nothing
        # left to execute
        yield self.node.qmemory.execute_program(program)
        self._gate_tuples_evaluated[-1].extend(self._local_gate_tuples_pending)
        # re-initialising self._local_gate_tuples_pending
        self._local_gate_tuples_pending = []
        self.send_signal(self.finished_time_slice_label)

        # TO DO: add entanglement swapping

    def raise_error_if_not_all_gates_executed(self):
        """
        To be run at the end of the protocol's run.
        """
        if self._gate_tuples_evaluated != self.gate_tuples:
            index_of_last_time_slice_evaluated = len(self._gate_tuples_evaluated) - 1
            index_of_final_time_slice = len(self.gate_tuples) - 1
            try:
                last_gate_tuple_evaluated = self._gate_tuples_evaluated[-1][-1]
            except IndexError:
                last_gate_tuple_evaluated = self._gate_tuples_evaluated[-1]
            raise UnfinishedQuantumCircuitError(
                "Not all operations in the quantum circuit have run. "
                "Evaluation stopped in time slice "
                f"{index_of_last_time_slice_evaluated} out of "
                f"{index_of_final_time_slice}. The last gate_tuple to "
                f"be evaluated was {last_gate_tuple_evaluated}. "
                "Overall, the gates evaluated were "
                f"{self._gate_tuples_evaluated} and the gate tuples "
                f"that SHOULD have been evaluated were "
                f"{self.gate_tuples}"
            )

    def run(self):
        if "physical_layer_protocol" not in self.subprotocols or not isinstance(
            self.subprotocols["physical_layer_protocol"], PhysicalLayerProtocol
        ):
            raise ValueError(
                f"{self.name} requires a physical layer protocol, "
                "which is a subclass of PhysicalLayerProtocol,"
                "to be added. This can be done with the "
                "add_phys_layer method"
            )

        self.subprotocols["physical_layer_protocol"].start()

        for time_slice in range(self._max_num_time_slices):
            yield self.await_signal(
                self.superprotocol, signal_label=self.start_time_slice_label
            )
            gate_tuples4time_slice = self.gate_tuples[time_slice]
            yield from self._run_time_slice(gate_tuples4time_slice)
            if time_slice < (self._max_num_time_slices - 1):
                # synchronise number of time slices in
                # self._gate_tuples_evaluated with the number of time slices in
                # self._gate_tuples
                self._gate_tuples_evaluated.append([])


class DQCMasterProtocol(LocalProtocol):
    """Protocol which executes a distributed quantum circuit.

    Parameters
    ----------
    partitioned_gates:  list of tuples.
        The gates implemented in the entire circuit. The tuples should be
        of the form:
        1) single-qubit gate: (gate_instr, qubit_index, node_name)
        2) local two-qubit gate: (gate_instr, qubit_index0, node0_name,
                                  qubit_index1, node1_name)
        3) two-qubit gate: (list of instructions or gate_instruction if
                            local, qubit_index0, node0_name, qubit_index1,
                            node1_name, scheme) where
                            list of instructions: list
                                list of same form as partitioned_gates
                                containing the local gates to be conducted
                                as part of the remote gate. Ie, for
                                remote-cnot this would contain the cnot
                                (but not the gates used for bsm or
                                 correction).
                                Note if this is given as empty list and
                                scheme = "tp" then it will just do a
                                teleportation
    physical_layer_protocol_class: class or None
        The class of the physical layer protocol. The choice must be callable
        with no explicit arguments. If None,
        :class:`~dqc_simulator.software.physical_layer.AbstractCentralSourceEntangleProtocol`
        is used.
    background_protocol_lookup : dict
        Lookup table with keys being subclasses of netsquid.nodes.node.Node
        and protocols being non-QPU subclasses of
        netsquid.protocols.protocol.Protocol. This is used to start all
        protocols that exist outside of time slice structure and must be
        initialised to make the network function.
    compiler_func: function, optional
        The compiler function to be used to split partitioned_gates into
        protocols by node and time-slice. If None (default), then
        :func: `~dqc_simulator.software.compilers.sort_greedily_by_node_and_time`
        is used.
    num_entanglement_attempts : int or None, optional
        The number of entanglement attempts for the physical layer to use
        before returning control to higher layers. This must be specified if
        the physical layer entanglement protocol used is not deterministic. The
        default is None (which assumes the physical layer is deterministic by
        default).
    name : str or None, optional. Name of protocol. If None, the name of the
        class is used [1]_.

    Attributes
    ----------
    physical_layer_protocol : :class:`~dqc_simulator.software.physical_layer.Base4PhysicalLayer
        The physical layer protocol used.

    Notes
    -----
    This protocol violates locality by splitting everything into time slices.
    Each time slice
    obeys locality but the entire network is instantaneously made aware when a
    time slice ends. This is intended to be an easier-to-implement abstraction
    of each node knowing the entire circuit and having a time slice
    counter of things it is allowed to do on that time slice before waiting a
    certain amount of time (the max time for operations on any QPU in the
    entire circuit) prior to evaluating the operations it has for that time slice.
    This construction is a compromise between retaining the
    user's ability to easily use and test their own almost arbitrary compilation
    schemes (global sequencing can be enforced using the time slices,
    simplifying the coding of compilers to the production of nested lists of
    gate tuples) and an easier to use framework that waits for resources to
    become available
    and greedily conducts operations as soon as possible, which can be achieved
    by putting everything in one time slice. A limitation of the framework is
    that a given QPU may not be able to do subprotocols for more than one remote
    gate at once even if there are enough resources to do so (I am essentially
    assuming that each QPU can be involved in only one remote gate per time
    slice).

    References
    ----------
    The parameters in which Ref. [1]_ was cited were inherited from
    :class:`~netsquid.protocols.protocol.Protocol` and the description
    used for those parameters was taken from the NetSquid documentation [1]_

    .. [1] https://netsquid.org/

    .. todo::

        Decide whether to use *args4physical_layer, **kwargs for physical layer
        to allow the physical layer protocol to be called with arguments
    """

    def __init__(
        self,
        partitioned_gates,
        nodes=None,
        physical_layer_protocol_class=None,
        background_protocol_lookup=None,
        compiler_func=None,
        num_entanglement_attempts=None,
        name=None,
        max_nodes=-1,
    ):
        super().__init__(nodes, name, max_nodes)
        self.start_time_slice_label = "START_TIME_SLICE"
        self.finished_time_slice_label = "FINISHED_TIME_SLICE"
        self.add_signal(self.start_time_slice_label)
        self.add_signal(self.finished_time_slice_label)
        self.partitioned_gates = partitioned_gates
        # TO DO: think about whether it would be better to instantiate default
        # arguments using the setter of a property

        # instantiating default arguments
        if physical_layer_protocol_class is None:
            physical_layer_protocol_class = AbstractCentralSourceEntangleProtocol
        else:
            # TO DO: raise error if calling this without arguments raises error
            physical_layer_protocol_class = physical_layer_protocol_class
        if compiler_func is None and len(self.nodes) <= 2:
            self.compiler_func = sort_greedily_by_node_and_time
        elif compiler_func is None and len(self.nodes) > 2:
            self.compiler_func = sort_many_qpus_greedily_by_node_and_time
        else:
            self.compiler_func = compiler_func

        self.background_protocol_lookup = background_protocol_lookup

        self.qpu_op_dict = self.compiler_func(self.partitioned_gates)
        all_nodes = self.nodes
        for node_name in self.qpu_op_dict:
            # I use qpu_op_dict rather than network.nodes here to avoid any
            # situations in which a QPU node has not been added to the network but
            # has instructions specified for it without an error being raised
            node = all_nodes[node_name]
            if isinstance(node.qmemory, QPU):
                # if the node has a QPU and that QPU has instructions to run at any
                # point
                qpu_protocol = InterpreterProtocol(
                    superprotocol=self,
                    gate_tuples=self.qpu_op_dict[node_name],
                    node=node,
                    name=f"InterpreterProtocol_{node.name}",
                )
                self.physical_layer_protocol = physical_layer_protocol_class(node=node)
                if not self.physical_layer_protocol.deterministic:
                    self.physical_layer_protocol.num_entanglement_attempts = (
                        num_entanglement_attempts
                    )

                qpu_protocol.add_phys_layer(self.physical_layer_protocol)
                self.add_subprotocol(qpu_protocol, name=f"{node_name}_OS")

    # =============================================================================
    #             else:
    #                 node_type = type(node)
    #                 background_protocol = self.background_protocol_lookup[node_type](
    #                                          node=node,
    #                                          name=f"{node_name}_protocol")
    #                 background_protocol.start()
    # =============================================================================

    def check_quantum_circuit_finished(self):
        """
        Should be called after the sim has finished running to check if the
        intended distributed quantum circuit actually ran all the way through.
        """
        for qpu_name in self.qpu_op_dict:
            self.subprotocols[f"{qpu_name}_OS"].raise_error_if_not_all_gates_executed()

    def run(self):
        for qpu_name in self.qpu_op_dict:
            self.subprotocols[f"{qpu_name}_OS"].start()
        # initialising dummy event expression
        dummy_entity = pydynaa.Entity()
        evtype_dummy = pydynaa.EventType("dummy_event", "dummy event")
        evexpr_dummy = pydynaa.EventExpression(
            source=dummy_entity, event_type=evtype_dummy
        )
        expr = evexpr_dummy
        dummy_entity._schedule_now(evtype_dummy)

        # finding max length of dictionary entry
        longest_list_in_qpu_op_dict = max(self.qpu_op_dict.values(), key=len)
        max_num_time_slices = len(longest_list_in_qpu_op_dict)
        for time_slice in range(max_num_time_slices):
            for qpu_name in self.qpu_op_dict:
                # if QPU still has instructions to carry out:
                if time_slice < len(self.qpu_op_dict[qpu_name]):
                    # strictly less than because python indexes from 0
                    # signalling subprotocols to start the next time slice
                    self.send_signal(self.start_time_slice_label)
                    # waiting on subprotocols to complete time slice
                    expr = expr & self.await_signal(
                        self.subprotocols[f"{qpu_name}_OS"],
                        self.finished_time_slice_label,
                    )
            yield expr
            # re-initialising
            expr = evexpr_dummy
            dummy_entity._schedule_now(evtype_dummy)

        self.check_quantum_circuit_finished()


# For backwards compatability
dqcMasterProtocol = DQCMasterProtocol


# Correcting occupies a qubit because that needs to be used as the control for
# a cnot etc but the bsm frees one if done to teleport a comm_qubit

# I am assuming:
# 1) The input list of gate tuples is loosely ordered like in a circuit diagram
# 2) That there is only one remote gate per node on a time slice. This is by
# construction and so is not really an assumption. All of that said, it is
# actually possible to have multiple remote gates on a node if the compiler
# allows it. In that case many of the safeguards here will not work and so you
# need the compiler to have properly accounted for resources.
# 3) For now, gates cannot be done in parallel with receiving entanglement. This
# is probably physically realistic anyway.
# 4) At the moment, unlike the CollComm protocol explored by the Santa Barbara
# group, everything is quite coupled to the comm-qubits. This is not completely
# general



"""
Distributed quantum computing compilers.

These compilers take a list of gate tuples, and output a time-sliced dictionary
of what happens on each individual QPU.

Notes
-----
Some compilation is also done in dqc_control.py. There common
subroutines used by Cat-comm and TP-comm are broken down into individual gates
and run on individual nodes. The role of these compilers is to break a
monolithic quantum circuit into the lower level commands used in
:mod: `~dqc_simulator.software.dqc_control`.
That said, very general compilers can be defined here. The main limitations are
that they use only intraprocessor quantum gates and the interprocessor
subroutines involved in Cat and TP-comm (at least for now) and that the backend
defined in :mod: `~dqc_simulator.software.dqc_control` may not be able to cope
with carrying out multiple inter-QPU gates on a given node within the same
time-slice. Each time slice is intended to be
greedily compiled and executed by functions in
:mod: `~dqc_simulator.software.dqc_control` but near-arbitrary compilation can
be done by careful choice of the time slices.
"""

# Note: some compilation is also done in dqc_control.py. There common
# subroutines used by Cat-comm and TP-comm are broken down into individual gates
# and run on individual nodes. The role of these compilers is to break a
# monolithic quantum circuit into the lower level commands used in dqc_control.py.
# That said, very general compilers can be defined here. The only limitation is
# that they use only intraprocessor quantum gates and the interprocessor
# subroutines involved in Cat and TP-comm

from functools import partial

import netsquid as ns
from netsquid.components import instructions as instr


class QpuOps:
    """
    Data structure for tracking the operations of all QPUs within a DQC.

    Intended to help users develop their own compilers.

    Attributes
    ----------
    ops : dict
        The operations to carry out on each QPU in a DQC.

    Notes
    -----
    The definitions of cat-comm and TP-comm subroutines work less well for
    many QPU systems (rather than 2 QPU systems) and so this is best used for
    compiling 2-QPU systems

    """

    def __init__(self):
        self.ops = {}

    # =============================================================================
    #         self.internode_primitives = {
    #             'request_ebit' : }
    # =============================================================================

    def add_empty_node_entry(self, node_name):
        self.ops[node_name] = [[]]

    def append_op2current_time_slice(self, node_name, op):
        self.ops[node_name][-1].append(op)

    def append_multiple_ops2current_time_slice(self, node_name, ops):
        self.ops[node_name][-1] = self.ops[node_name][-1] + ops

    def add_op(self, node_name, op):
        if node_name not in self.ops:
            self.add_empty_node_entry(node_name)
        self.append_op2current_time_slice(node_name, op)

    def add_time_slice(self, node_name):
        self.ops[node_name].append([])

    def pad_num_time_slices_until_matching(
        self, node_with_fewer_time_slices, node_with_more_time_slices
    ):
        length_diff = len(self.ops[node_with_more_time_slices]) - len(
            self.ops[node_with_fewer_time_slices]
        )
        # adding empty time slices to make lengths match (enforcing
        # scheduling)
        start = len(self.ops[node_with_fewer_time_slices])
        end = start + length_diff
        for ii in range(start, end):
            self.add_time_slice(node_with_fewer_time_slices)

    def make_num_time_slices_match(self, node0_name, node1_name):
        if len(self.ops[node0_name]) > len(self.ops[node1_name]):
            self.pad_num_time_slices_until_matching(node1_name, node0_name)
        elif len(self.ops[node0_name]) < len(self.ops[node1_name]):
            self.pad_num_time_slices_until_matching(node0_name, node1_name)

    def check_if_all_nodes_have_empty_trailing_time_slice(self):
        """


        Returns
        -------
        all_nodes_have_empty_trailing_time_slice : bool
            True if all nodes have an empty trailing time slice and false
            otherwise.
        """
        all_nodes_have_empty_trailing_time_slice = True
        for node_key in self.ops:
            if self.ops[node_key][-1]:
                # if the last time slice is non-empty:
                all_nodes_have_empty_trailing_time_slice = False
        return all_nodes_have_empty_trailing_time_slice

    def remove_empty_trailing_time_slices(self):
        """
        Remove empty trailing time slices if they appear for all QPUs.

        The constant number of time slices for all nodes is retained.
        """
        need2remove_trailing_time_slices = (
            self.check_if_all_nodes_have_empty_trailing_time_slice()
        )
        if need2remove_trailing_time_slices:
            # removing last time slice from each node
            for node_key in self.ops:
                del self.ops[node_key][-1]

    # =============================================================================
    #     def request_ebit(self, node0_name, node1_name):
    #         """
    #         Request ebit (entangled pair) between two QPU nodes.
    #
    #         Parameters
    #         ----------
    #         node0_name, node1_name : str
    #             The names of the nodes to receive each half of the ebit.
    #         """
    #         node0_ops = [()]
    # =============================================================================

    def distribute_ebit(self, gate_tuple):
        # It doesn't matter here which node is the client and which is the
        # server
        node0_name = gate_tuple[0]
        node1_name = gate_tuple[1]
        node0_op = (node1_name, "client", "distribute_ebit")
        node1_op = (node0_name, "server", "distribute_ebit")
        self.append_op2current_time_slice(node0_name, node0_op)
        self.append_op2current_time_slice(node1_name, node1_op)

    def cat_comm(
        self, gate_instructions, qubit_index0, qubit_index1, node0_name, node1_name
    ):
        # =============================================================================
        #         node0_op = [(qubit_index0, node1_name, "cat", "entangle")]
        #         node1_op = [(node0_name, "cat", "correct")]
        #         self.append_op2current_time_slice(node0_name, node0_op)
        #         self.append_op2current_time_slice(node1_name, node1_op)
        #         self.add_time_slice(node0_name)
        #         self.add_time_slice(node1_name)
        #         node0_ops = [(qubit_index0, node1_name, "cat", "disentangle_end")]
        #         node1_ops = ([gate_instructions] +
        #                     [(qubit_index1, node0_name, "cat", "disentangle_start")])
        #         self.append_multiple_ops2current_time_slice(node0_name, node0_ops)
        #         self.append_multiple_ops2current_time_slice(node1_name, node1_ops)
        #         self.add_time_slice(node0_name)
        #         self.add_time_slice(node1_name)
        # =============================================================================
        node0_ops = [(qubit_index0, node1_name, "cat", "entangle")] + [
            (qubit_index0, node1_name, "cat", "disentangle_end")
        ]
        node1_ops = (
            [(node0_name, "cat", "correct")]
            + gate_instructions
            + [(qubit_index1, node0_name, "cat", "disentangle_start")]
        )
        self.append_multiple_ops2current_time_slice(node0_name, node0_ops)
        self.append_multiple_ops2current_time_slice(node1_name, node1_ops)

    # =============================================================================
    #         self.add_time_slice(node0_name)
    #         self.add_time_slice(node1_name)
    # =============================================================================

    def one_tp(self, gate_instructions, qubit_index0, node0_name, node1_name):
        """
        Add gate tuples for 1tp remote gate, in which a processing qubit
        is teleported to an automatically assigned comm-qubit in another QPU
        and then the gate or gates are done locally by the QPU.

        No teleportation is done back to the original QPU to free up the
        comm-qubit used.
        """
        node0_ops = [(qubit_index0, node1_name, "tp", "bsm")]
        node1_ops = [(node0_name, "tp", "correct")] + gate_instructions
        self.append_multiple_ops2current_time_slice(node0_name, node0_ops)
        self.append_multiple_ops2current_time_slice(node1_name, node1_ops)

    def teleport_only(
        self,
        gate_instructions,
        proc_qubit2teleport_from,
        comm_qubit2teleport_to,
        node0_name,
        node1_name,
    ):
        """
        Add gate tuples for a teleportation.

        No subsequent gate is done automatically.

        Parameters
        ----------
        gate_instructions : empty list
            Must be empty list, and is included only for input checking
        proc_qubit2teleport_from : int
            The index of the processing qubit to teleport from (the control
            qubit in the remote gate).
        comm_qubit2teleport_to : int
            The index of the comm-qubit to be teleported to.
        node0_name : str
            The name of the QPU node to be teleported from.
        node1_name : str
            The name of the QPU node to be teleported to.

        .. todo::

            Deprecate the type check here, by changing the way gate tuples
            with four or more elements are interpreted. Length alone is not
            a stable interpretation method.
        """
        if gate_instructions != []:
            # TO DO: deprecate this type check by changing the way that
            # interpretation of gate_tuples is done
            raise ValueError(
                "gate_instructions must be empty list because "
                "teleport_only involves only teleportation "
            )
        node0_ops = [(proc_qubit2teleport_from, node1_name, "tp", "bsm")]
        node1_ops = [(comm_qubit2teleport_to, node0_name, "tp", "correct_manually")]
        self.append_multiple_ops2current_time_slice(node0_name, node0_ops)
        self.append_multiple_ops2current_time_slice(node1_name, node1_ops)

    def free_comm_qubit_with_tele(
        self, qubit_index0, qubit_index1, node0_name, node1_name
    ):
        # TO DO: potentially deprecate in favour of tele2data_qubit, however
        # this will require changes to dqc_simulator.software.dqc_control
        node0_ops = [(qubit_index0, node1_name, "tp", "bsm")]
        node1_ops = [
            (node0_name, "tp", "correct4tele_only"),
            (instr.INSTR_SWAP, -1, qubit_index1),
        ]
        self.append_multiple_ops2current_time_slice(node0_name, node0_ops)
        self.append_multiple_ops2current_time_slice(node1_name, node1_ops)

    def tele2data_qubit(self, qubit_index0, qubit_index1, node0_name, node1_name):
        node0_ops = [(qubit_index0, node1_name, "tp", "bsm")]
        node1_ops = [(qubit_index1, node0_name, "tp", "swap_then_correct")]
        self.append_multiple_ops2current_time_slice(node0_name, node0_ops)
        self.append_multiple_ops2current_time_slice(node1_name, node1_ops)

    def tp_safe(
        self, gate_instructions, qubit_index0, qubit_index1, node0_name, node1_name
    ):
        """Specifying gate tuples needed to implement remote gate using 1tp,
        then teleports back to original qubit on the QPU to free up comm-qubits.

        In the ideal case, the qubits would be left in the state they would
        have been had it been possible to do a local gate between them (ie,
        all quantum information is stored in processing qubits).
        """
        # For remote gate:
        self.one_tp(gate_instructions, qubit_index0, node0_name, node1_name)
        self.add_time_slice(node0_name)
        self.add_time_slice(node1_name)
        # for teleportation back:
        self.tele2data_qubit(-1, qubit_index0, node1_name, node0_name)
        self.add_time_slice(node0_name)
        self.add_time_slice(node1_name)
        # implements tp remote gate then teleports back to original node
        # to free up comm-qubit.

    # =============================================================================
    #         #For remote gate:
    #         self.one_tp(gate_instructions, qubit_index0, node0_name,
    #                       node1_name)
    #         self.add_time_slice(node0_name)
    #         self.add_time_slice(node1_name)
    #         #For teleportation back:
    #         self.free_comm_qubit_with_tele(-1, qubit_index0, node1_name,
    #                                        node0_name)
    #         self.add_time_slice(node0_name)
    #         self.add_time_slice(node1_name)
    # =============================================================================

    # =============================================================================
    #     def tp_block(self, gate_instructions, qubit_index0, qubit_index1,
    #                  node0_name, node1_name):
    #         #Very similar to TP-safe except that the SWAP happens before the
    #         #correction and the measurement dependent gates act on the data qubit
    #         #(see figure 2b of published version of AutoComm paper)
    #         #For remote gate:
    #         self.one_tp(gate_instructions, qubit_index0, node0_name,
    #                       node1_name)
    #         self.add_time_slice(node0_name)
    #         self.add_time_slice(node1_name)
    #         #for teleportation back:
    #         self.tele2data_qubit(-1, qubit_index0, node1_name, node0_name)
    #         self.add_time_slice(node0_name)
    #         self.add_time_slice(node1_name)
    # =============================================================================

    def apply_remote_gate(
        self,
        scheme,
        gate_instructions,
        qubit_index0,
        qubit_index1,
        node0_name,
        node1_name,
    ):
        # the use of functools.partial in the following just bakes in the
        # arguments and allows a single call of the dict to be used despite
        # the arguments being different
        comm_schemes = {
            "cat": partial(
                self.cat_comm,
                gate_instructions,
                qubit_index0,
                qubit_index1,
                node0_name,
                node1_name,
            ),
            "1tp": partial(
                self.one_tp, gate_instructions, qubit_index0, node0_name, node1_name
            ),
            "teleport_only": partial(
                self.teleport_only,
                gate_instructions,
                qubit_index0,
                qubit_index1,
                node0_name,
                node1_name,
            ),
            "free_comm_qubit_with_tele": partial(
                self.free_comm_qubit_with_tele,
                qubit_index0,
                qubit_index1,
                node0_name,
                node1_name,
            ),
            "tp_safe": partial(
                self.tp_safe,
                gate_instructions,
                qubit_index0,
                qubit_index1,
                node0_name,
                node1_name,
            ),
        }

        comm_schemes[scheme]()


# helper functions:


def find_qubit_node_pairs(partitioned_gates):
    """

    Parameters
    ----------
    partitioned_gates : list of tuples
        List of tuples with gate information including type of gate, qubit
        indices acted on, node names acted on, and (for remote gates) a string
        indicating the scheme (here this should be some placeholder string
        as the scheme is not yet decided).

    Returns
    -------
    qubit_node_pairs : dict
        All of the gates associated with a qubit-node pair. Qubit-node pairs
        are referred to by a key of the form (qubit0_index, node0_name,
        node1_name), where the first two
        entries define the qubit and the latter the node. The remote gates for
        each qubit-node pair are lists of the form [gate, ID] where gate is
        the gate tuple form partitioned gates and ID
    """

    def _add_entry(a_dict, key, value):
        if key in a_dict:
            a_dict[key].append(value)
        else:
            a_dict[key] = [value]

    qubit_node_pairs = {}
    for ii, gate in enumerate(
        partitioned_gates
    ):  # you will also iterate over gates of different
        # type in linear merge but I'm not sure this can be
        # avoided
        if len(gate) > 3 and gate[2] != gate[4]:
            # if two-qubit remote gate (node names not the same):
            ctrl_qubit = gate[1]
            ctrl_node = gate[2]
            target_qubit = gate[3]
            target_node = gate[4]
            key = (ctrl_qubit, ctrl_node, target_node)
            # first two entries of key needed to fully specify ctrl qubit
            _add_entry(qubit_node_pairs, key, (gate, ii))
            # index ii needed so that duplicate gates can be identified later
            # and their position in partitioned_gates can be ascertained
            key = (target_qubit, target_node, ctrl_node)
            _add_entry(qubit_node_pairs, key, (gate, ii))
    return qubit_node_pairs


def find_node_node_pairs(partitioned_gates):
    node_node_pairs = {}
    for ii, gate in enumerate(partitioned_gates):
        if len(gate) > 3 and gate[2] != gate[4]:
            # if two-qubit remote gate (node names not the same):
            ctrl_node = gate[2]
            target_node = gate[4]
            key = (ctrl_node, target_node)
            key = tuple(sorted(key))  # puts key in intuitive ordering
            # (eg, node_0, node_1, ... rather than
            # node_1, node_0, ...)
            if key in node_node_pairs:
                node_node_pairs[key].append((gate, ii))
                # index ii needed so that the position in partitioned_gates can
                # be ascertained later, during scheduling, to avoid illegal
                # commutation
            elif (key[1], key[0]) in node_node_pairs:
                node_node_pairs[(key[1], key[0])].append((gate, ii))
            else:
                node_node_pairs[key] = [(gate, ii)]
    return node_node_pairs


def find_pairs(partitioned_gates, pair_type):
    """
    Parameters
    ----------
    partitioned_gates : list of tuples
        List of tuples with gate information including type of gate, qubit
        indices acted on, node names acted on, and (for remote gates) a string
        indicating the scheme (here this should be some placeholder string
        as the scheme is not yet decided).
    pair_type : str
        Whether to return 'qubit_node' or 'node_node' pairs.

    Returns
    -------
    pairs : dict
        The remote gates for each qubit-node or node-node pair

    """
    if pair_type == "qubit_node":
        pairs = find_qubit_node_pairs(partitioned_gates)
    elif pair_type == "node_node":
        pairs = find_node_node_pairs(partitioned_gates)
    return pairs


def order_pairs(pairs):
    """
    Sorts dict of the remote gates associated with different qubit-node or
    node-node pairs in descending order

    Parameters
    ----------
    pairs : dict
        Sorts dict of the remote gates associated with different qubit-node or
        node-node pairs

    Returns
    -------
    dict
        The pairs ordered from those with the most remote gates to those with
        the least.

    """
    return dict(sorted(pairs.items(), key=lambda items: len(items[1]), reverse=True))


# =============================================================================
# def something_btwn_gates(all_gates, gate1_info, gate2_info):
#     if element1
# =============================================================================


def get_qubit_id(gate):
    """
    Collecting info needed to uniquely ID the qubits of a gate within a
    partitioned circuit

    Parameters
    ----------
    gate : TYPE
        DESCRIPTION.

    Returns
    -------
    qubit0 : TYPE
        DESCRIPTION.
    qubit1 : TYPE
        DESCRIPTION.

    """
    # TO DO: convert this function to handle multi-qubit gate using the pattern
    # of indices (would need to return a single tuple which can be unpacked)
    if len(gate) == 3:
        qubit_index0 = gate[1]
        qubit_index1 = None
        node_index0 = gate[2]
        node_index1 = None
        qubit1 = (None, None)
    elif len(gate) > 3:
        qubit_index0 = gate[1]
        qubit_index1 = gate[3]
        node_index0 = gate[2]
        node_index1 = gate[4]
        qubit1 = (qubit_index1, node_index1)  # fully defining 2nd qubit
    qubit0 = (qubit_index0, node_index0)
    return qubit0, qubit1


def shared_qubit(gate0, gate1):
    gate0qubit0, gate0qubit1 = get_qubit_id(gate0)
    gate1qubit0, gate1qubit1 = get_qubit_id(gate1)
    gate0shares_gate1qubit0 = (
        gate0qubit0 == gate1qubit0 or gate0qubit0 == gate1qubit1
    ) and None not in gate0qubit0
    gate0shares_gate1qubit1 = (
        gate0qubit1 == gate1qubit0 or gate0qubit1 == gate1qubit1
    ) and None not in gate0qubit1
    return gate0shares_gate1qubit0 or gate0shares_gate1qubit1


def find_consecutive_gate(partitioned_gates, gate, gate_position):
    """

    Finds the next gate which has shared qubits with the specified `gate`.

    Otherwise, the next gate in partitioned gates may be completely unrelated
    to the gate of interest with no need to maintain their ordering. They
    may even occur on completely different nodes or pairs of nodes.

    Parameters
    ----------
    partitioned_gates : list of tuples
        DESCRIPTION.
    gate: tuple
        A gate.
    gate_position : int
        The positional index of the gate in partitioned gates.

    Returns
    -------
    nxt_gate : tuple
        The gate consecutive to `gate`.
    """
    # TO DO: replace following with generator for memory efficiency to avoid
    # storing whole sublist
    unvisited_partitioned_gates = partitioned_gates[gate_position + 1 :]
    nxt_gate = None
    for gate2check in unvisited_partitioned_gates:
        # =============================================================================
        #         print(f'gate of interest is {gate}')
        #         print(f'gate to check is {gate2check}')
        #         print(f'they are consecutive: {shared_qubit(gate, gate2check)}')
        # =============================================================================
        if shared_qubit(gate, gate2check):
            nxt_gate = gate2check
            # =============================================================================
            #             print(f'consecutive gate is {gate2check}')
            # =============================================================================
            break
    return nxt_gate


# TO DO: think about whether the following function should be internal to
# aggregate_comms (in which case filtered_remote_gates would be qubit_node_pairs)
def find_consecutive_remote_gates(partitioned_gates, filtered_remote_gates):
    """
    Finds the consecutive gates in a filtered set of remote gates (eg, those
    in the same qubit-node or node-node pair)

    Parameters
    ----------
    partitioned_gates : list of tuples
        All gates in quantum circuit (partitioned into different nodes)
    filtered_remote_gates : list of tuples of form
                            (gate, postional index of gate within partitioned gates)
        The filtered list of remote gates and their positional indices wrt
        the overall list of partitioned gates.


    Returns
    -------
    burst_comm_blocks : list of lists
        The identified burst communication blocks as found in the first step of
        the autocomm communication aggregation process (identifying blocks)
        without circuit restructuring.

    """
    current_filtered_gateNindex = filtered_remote_gates[0]
    current_filtered_gate = current_filtered_gateNindex[0]
    current_filtered_gate_position = current_filtered_gateNindex[1]
    burst_comm_blocks = [[current_filtered_gateNindex]]
    unvisited_gates = filtered_remote_gates[1:]
    for ii, nxt_remote_gateNindex in enumerate(unvisited_gates, start=1):
        nxt_filtered_gate = nxt_remote_gateNindex[0]
        nxt_filtered_gate_position = nxt_remote_gateNindex[1]
        nxt_partitioned_gate = find_consecutive_gate(
            partitioned_gates, current_filtered_gate, current_filtered_gate_position
        )
        if nxt_partitioned_gate == nxt_filtered_gate:
            # if the remote gates from the filtered set ARE consecutive:
            burst_comm_blocks[-1].append(nxt_remote_gateNindex)
        else:
            # if the remote gates from the filtered set are NOT consecutive:
            burst_comm_blocks.append([nxt_remote_gateNindex])
        current_filtered_gate = nxt_filtered_gate
        current_filtered_gate_position = nxt_filtered_gate_position
    # removing trailing empty sublist
    if not burst_comm_blocks[-1]:
        # if final element empty:
        del burst_comm_blocks[-1]
    return burst_comm_blocks
    # TO DO: figure out why this is not giving desired output


# =============================================================================
# def aggregate_comms(partitioned_gates):
#     qubit_node_pairs = find_pairs(partitioned_gates, 'qubit_node')
#     #ordering pairs by number of associated remote gates (largest first):
#     qubit_node_pairs = order_pairs(qubit_node_pairs)
#     for qubit_node_pair in qubit_node_pairs:
#         burst_comm_blocks = find_consecutive_remote_gates(partitioned_gates,
#                                                           qubit_node_pairs)
#         #linear fusion here on burst_comm_blocks
# # =============================================================================
# #         for remote_gateNindex in qubit_node_pair:
# #             pos_in_partitioned_gates = remote_gateNindex[1]
# # =============================================================================
# #TO DO (longer term): generalise aggregate_comms to work with node-node pairs as well
# =============================================================================


def _find_all_nodes(partitioned_gates):
    qpu_node_names = []
    for gate_tuple in partitioned_gates:
        for node_name in gate_tuple[2::2]:
            if node_name not in qpu_node_names:
                qpu_node_names.append(node_name)
    qpu_node_names.sort()
    return qpu_node_names


# compilers
# ---------


def sort_greedily_by_node_and_time(
    partitioned_gates, new_time_slice4each_remote_gate=False
):
    """
    Distributes the circuit between nodes and splits into explicit time-slices
    (rows in output array). Initialisation of qubits must be specified as
    an instruction in the partitioned_gates input.

    Parameters
    ----------
    partitioned_gates:  list of tuples
        The gates in the entire circuit. The tuples should be of the form:
        1) single-qubit gate: (gate_instr, qubit, node_name)
        2) two-qubit gate: (list of instructions or gate_instruction or
                            instruction_tuple if local,
                            qubit0, node0_name, qubit1, node1_name, scheme if remote)
                            #can later extend this to multi-qubit gates
                            #keeping scheme as last element
                            list of instructions: list
                                list of same form as partitioned_gates containing
                                the local gates to be conducted as part of the
                                remote gate. Ie, for remote-cnot this would
                                contain the cnot (but not the gates used for
                                 bsm or correction).Note if this is given as
                                empty list and scheme = "1tp" then it will
                                just do a teleportation
                            gate_instruction : :class:`~netsquid.components.instructions.Instruction`
                                The gate instruction for the target gate
                            instruction_tuple : tuple
                                Tuple of form (gate_instruction, op), where
                                op is the operatution used to perform the
                                gate. This form is useful if you want to give
                                several gates the same
                                `netsquid.components.qprocessor.PhysicalInstruction`
                            qubit{ii}: int
                                The qubit index
                            node_name: str
                            The name of the relevant node
                            scheme: str, optional
                                Only needed for remote gates
    new_time_slice4each_remote_gate : bool, optional
        Whether to create a new time slice for all QPUs for every remote gates.
        This is a simplistic way of avoiding scheduling conflicts when more
        than 2 QPUs are used. The drawback is that the latency will be higher.

    Notes
    -----
    The scheduling here works less well for more than two QPUs and may lead to
    errors.

    Returns
    -------
    dict
        Dictionary with term for each node which is an list of different
        time slices (each time slice is a list within the list)
    """
    qpu_ops = QpuOps()

    for gate_tuple in partitioned_gates:
        if len(gate_tuple) == 3 and gate_tuple[-1] == "distribute_ebit":
            node0_name = gate_tuple[0]
            node1_name = gate_tuple[1]
            if node0_name not in qpu_ops.ops:
                qpu_ops.add_empty_node_entry(node0_name)
            if node1_name not in qpu_ops.ops:
                qpu_ops.add_empty_node_entry(node1_name)

            qpu_ops.make_num_time_slices_match(node0_name, node1_name)
            qpu_ops.distribute_ebit(gate_tuple)
        elif len(gate_tuple) == 3:  # if single-qubit gate:
            gate_instr = gate_tuple[0]
            qubit_index = gate_tuple[1]
            node_name = gate_tuple[2]
            op = (gate_instr, qubit_index)
            qpu_ops.add_op(node_name, op)
        elif len(gate_tuple) == 4:  # if single-qubit_subroutine (gate with some
            # additional behaviour, eg measurement with
            # result saved):
            gate_instr = gate_tuple[0]
            qubit_index = gate_tuple[1]
            node_name = gate_tuple[2]
            subroutine_type = gate_tuple[3]
            op = (gate_instr, qubit_index, subroutine_type)
            qpu_ops.add_op(node_name, op)
        elif len(gate_tuple) > 4:  # if multi-qubit gate:
            qubit_index0 = gate_tuple[1]
            node0_name = gate_tuple[2]
            qubit_index1 = gate_tuple[3]
            node1_name = gate_tuple[4]
            if node0_name == node1_name:  # if local:
                gate_instr = gate_tuple[0]
                node_name = node0_name
                op = (gate_instr, qubit_index0, qubit_index1)
                qpu_ops.add_op(node_name, op)
            else:  # if remote gate:
                scheme = gate_tuple[-1]
                gate_instructions = gate_tuple[0]
                if isinstance(
                    gate_instructions, ns.components.instructions.IGate
                ) or isinstance(gate_instructions, tuple):
                    # putting into correct form to use the comm-qubit index
                    # as the control (defers exact index choice to
                    # InterpreterProtocol in
                    # dqc_simulator.software.dqc_control.py):
                    gate_instructions = [(gate_instructions, -1, qubit_index1)]
                if node0_name not in qpu_ops.ops:
                    qpu_ops.add_empty_node_entry(node0_name)
                if node1_name not in qpu_ops.ops:
                    qpu_ops.add_empty_node_entry(node1_name)

                qpu_ops.make_num_time_slices_match(node0_name, node1_name)
                qpu_ops.apply_remote_gate(
                    scheme,
                    gate_instructions,
                    qubit_index0,
                    qubit_index1,
                    node0_name,
                    node1_name,
                )
                # Time slices are added in the above line as part of the QpuOps
                # methods corresponding to specific remote gate schemes.

                if new_time_slice4each_remote_gate:
                    qpu_node_names = _find_all_nodes(partitioned_gates)
                    for node_name in qpu_node_names:
                        qpu_ops.add_time_slice(node_name)
    qpu_ops.remove_empty_trailing_time_slices()
    return qpu_ops.ops


def sort_many_qpus_greedily_by_node_and_time(partitioned_gates):
    """
    Distributes the circuit between QPU nodes and splits into explicit time-slices
    (rows in output array). Initialisation of qubits must be specified as
    an instruction in the partitioned_gates input. Unlike
    `sort_greedily_by_node_and_time` with the default settings, this compiler
    is designed to avoid scheduling conflicts when more than 2 QPUs are used.
    The drawback is that the latency will be higher because a new time slice is
    created every time a remote gate occurs.

    Parameters
    ----------
    partitioned_gates:  list of tuples
        The gates in the entire circuit. The tuples should be of the form:
        1) single-qubit gate: (gate_instr, qubit, node_name)
        2) two-qubit gate: (list of instructions or gate_instruction or
                            instruction_tuple if local,
                            qubit0, node0_name, qubit1, node1_name, scheme if remote)
                            #can later extend this to multi-qubit gates
                            #keeping scheme as last element
                            list of instructions: list
                                list of same form as partitioned_gates containing
                                the local gates to be conducted as part of the
                                remote gate. Ie, for remote-cnot this would
                                contain the cnot (but not the gates used for
                                 bsm or correction).Note if this is given as
                                empty list and scheme = "tp" then it will
                                just do a teleportation
                            gate_instruction : :class:`~netsquid.components.instructions.Instruction`
                                The gate instruction for the target gate
                            instruction_tuple : tuple
                                Tuple of form (gate_instruction, op), where
                                op is the operatution used to perform the
                                gate. This form is useful if you want to give
                                several gates the same
                                `netsquid.components.qprocessor.PhysicalInstruction`
                            qubit{ii}: int
                                The qubit index
                            node_name: str
                            The name of the relevant node
                            scheme: str, optional
                                Only needed for remote gates

    Returns
    -------
    dict
        Dictionary with term for each node which is an list of different
        time slices (each time slice is a list within the list)
    """
    return sort_greedily_by_node_and_time(
        partitioned_gates, new_time_slice4each_remote_gate=True
    )
    # The second option is the point of this wrapper.


# I think rather than compiling dqc_circuit you need to pre-process once more
# and add initialisation in that stage if you want to use existing machinery
# if not it needs updated to use instance of DqcCircuit

# For communication fusion blocks then you should be inputting a block of
# different gates sandwhiched by remote gate primitives into
# HandleCommBlockForOneNodeProtocol



# -*- coding: utf-8 -*-
# =============================================================================
#
# Created on Fri Sep 13 09:29:30 2024
#
# @author: kenny
# =============================================================================
"""
Protocols for the physical layer similar to that defined in the Wehner network
stack.

Notes
-----
The original [1]_ and revised [2]_ papers on the Wehner network stack give
further details.

References
----------
.. [1] A. Dahlberg, et al., A link layer protocol for quantum networks, in:
       Proceedings of the ACM Special Interest Group on Data Communication,
       ACM (2019)

.. [2] M. Pompili, et al., Experimental demonstration of entanglement delivery
       using a quantum network stack, npj Quantum Information, 8 (2022)


.. todo::

    Decide if this module should be called link_and_physical_layer to
    reflect the fact that the functionality of the physical layer is
    encapsulated in a method of the base class for the physical layer. Also,
    need to condiser using ABC things because you want to make sure that
    subclasses have certain behaviour.

"""
# TO DO: think about if this module should be called link_and_physical_layer to
# reflect the fact that the functionality of the physical layer is encapsulated
# in the base class for the physical layer.

from netsquid.components import instructions as instr
from netsquid.components import QuantumProgram
from netsquid.protocols import NodeProtocol

# =============================================================================
# from netsquid_trappedions import EmitProg #EmitProg not actually specific
#                                           #to trapped ions.
# =============================================================================
from netsquid.qubits import ketstates as ks


# TO DO: think about enforcing some expected inheritance behaviour, eg with
# properties


class PhysicalLayerProtocol(NodeProtocol):
    """
    An abstract base class for physical layer protocols.

    This contains everything that all physical layer protocols should use,
    including the link layer itself, which is implemented as the
    `run_link_layer` subgenerator method. This method should be called in the
    run method of all physical layer protocols, typically prior to anything
    else other than the beginning of the while loop needed for all of the
    :class:`~pydynaa.core.EventExpression` subgenerators.

    Parameters
    ----------
    node : :class:`~netsquid.nodes.node.Node` or None, optional
        Node this protocol runs on. If None, a node should be set later before
        starting this protocol. [1]_
    name : str or None, optional
        Name of protocol. If None, the name of the class is used.
    role : str or None, optional
        Whether the QPU is a 'client' (initiating the handshake) or a
        'server' (responding to the handshake). If None, must be overwritten
        prior to the start of the protocol, eg in the link layer protocol.
    other_node_name : str or None, optional. [1]_
        The name of the other QPU node involved in the entangled link. If
        None, must be overwritten prior to the start of the protocol, eg in the
        link_layer_protocol.
    comm_qubit_indices = list of int or None, optional
        The index of the memory position in which the comm-qubit that should
        host the entanglement is being held.
    ready4ent = bool, optional
        Whether the QPU is ready or not for entanglement

    Attributes
    ----------
    superprotocol : :class:`netsquid.protocols.nodeprotocols.NodeProtocol`
        The superprotocol for this protocol. This protocol will be
        executed as a subprotocol of the superprotocol. The value of this
        attribute should be overwritten by the superprotocol.
    num_entanglements2generate : int
        The number of instances of the requested entanglement to be specified.
        For many subclasses this will be fixed but in others it will be
        variable and will be overwritten by the link layer.
    entanglement_type2generate : str
        The type of entanglement to be used. For many subclasses this will be
        fixed but in others it will be variable and will be overwritten by the
        link layer.
    ent_ready_label : str
        The label to use to signal the successful distribution of
        entanglement.
    ent_failed_label : str
        The label to use to signal that entanglement distribution has
        failed

    Notes
    -----
    Although num_entanglements2generate and entanglement_type2generate to
    generate will not be used in many subclasses of this protocol, they are
    useful for the user to be able to find.

    References
    ----------
    The parameters in which Ref. [3]_ was cited were inherited from
    :class:`~netsquid.protocols.nodeprotocols.NodeProtocol` and the description
    used for those parameters was taken from the NetSquid documentation [3]_

    .. [3] https://netsquid.org/

    .. todo::

        Think about whether to set role, other_node_name, comm_qubit_indices,
        and read4ent as attributes or properties only, not parameters, as they
        should be set by protocols at higher layers. The advantage of this is
        to avoid confusion when a user sets values for these parameters
        and they are then overwritten. However, the parameter formulation does
        make instantiating the physical layer within higher layers easier.
    """

    def __init__(
        self,
        node=None,
        name=None,
        role=None,
        other_node_name=None,
        comm_qubit_indices=None,
        ready4ent=True,
    ):
        super().__init__(node=node, name=name)
        self.superprotocol = None  # should be overwritten by higher-layer
        # protocols. superprotocol will typically be
        # InterpreterProtocol from dqc_control module
        # and set in DQCMasterProtocol
        self.other_node_name = other_node_name
        self.comm_qubit_indices = comm_qubit_indices
        self.ready4ent = ready4ent
        # The following attributes will need to be overwritten in some
        # subclasses
        self.num_entanglements2generate = 1
        self.entanglement_type2generate = None
        self._deterministic = None  # made into read-only property below
        # the following two handshake labels are not strictly necessary because
        # any input message to the port would do the same thing. However, they
        # make what is going on clearer and facilitate debugging.
        self.handshake_ready_label = "READY4ENT"
        self.handshake_not_ready_label = "NOT_READY4ENT"
        self.ent_ready_label = "ENT_READY"
        self.ent_failed_label = "ENT_FAILED"
        self.ent_request_label = "ENT_REQUEST"
        self.add_signal(self.ent_ready_label)
        self.add_signal(self.ent_failed_label)
        self.add_signal(self.ent_request_label)
        # The value of the next four attributes will be overwritten in the
        # run method, which is the first point at which the node attribute must
        # have a value other than None
        self.classical_connection_port_name = None
        self.classical_conn_port = None
        self.entangling_connection_port_name = None
        self.ent_conn_port = None
        # TO DO: think about if you should declare these in InterpreterProtocol, which
        # you will need to do anyway, and then send them here with the other
        # attributes

    @property
    def deterministic(self):
        """
        Read-only bool property indicating whether the entanglement is
        deterministic or not

        Must be overwritten when subclassing by setting the value of
        self._deterministic in the __init__ method
        """
        if self._deterministic is None:
            raise NotImplementedError(
                "Must set value of read-only "
                "deterministic property by setting a "
                "value for self._deterministic in the"
                "__init__ method"
            )
        if not isinstance(self._deterministic, bool):
            raise TypeError(
                "The deterministic property must have type bool. "
                f"The currently set value ({self._deterministic})"
                "is not of type bool."
            )
        return self._deterministic

    @deterministic.setter
    def deterministic(self, value):
        raise TypeError(
            "deterministic is a read-only property - its valuecannot be set by the user"
        )

    def run_link_layer(self):
        """
        Handles entanglement requests from higher-level protocols.

        This is a subgenerator, allowing the run method to call it and wait on
        :class:`~pydynaa.core.EventExpression`s that this subgenerator waits
        on. It waits on a signal from higher-level protocols and changes this
        protocol's instance attributes to reflect information in that signal.
        It must be prepended by `yield from` when calling.

        Yields
        ------
        :class:`~pydynaa.core.EventExpression`
            Sends :class:`~pydynaa.core.EventExpression`s to the run method,
            causing it to wait on signals.
        """
        yield self.await_signal(self.superprotocol, signal_label=self.ent_request_label)
        # the following could be replaced with any desired specs (including
        # a tuple of them). TO DO: think about whether you want to have more
        # specs (eg, entanglement fidelity like in Wehner stack papers).
        # For now, I'll keep it simple
        signal_results = self.superprotocol.get_signal_result(self.ent_request_label)
        # updating relevant attributes
        self.role = signal_results[0]
        self.other_node_name = signal_results[1]
        self.comm_qubit_indices = signal_results[2]
        self.num_entanglements2generate = signal_results[3]
        self.entanglement_type2generate = signal_results[4]
        # the following port is for an extra classical connection used only for
        # the handshake (to avoid messages ending up in wrong protocols)
        self.classical_connection_port_name = self.node.connection_port_name(
            self.other_node_name, label="extra_classical"
        )
        self.classical_conn_port = self.node.ports[self.classical_connection_port_name]
        self.entangling_connection_port_name = self.node.connection_port_name(
            self.other_node_name, label="entangling"
        )
        self.ent_conn_port = self.node.ports[self.entangling_connection_port_name]

    def signal_outcome(self, ent_successful):
        """
        Signals entanglement generation outcome to
        QpuManagementProtocol/HandleCommBlockForOneNodeProtocol.

        Parameters
        ----------
        ent_successful : bool
            True if entanglement has been successfully distributed.
            False otherwise.

        .. todo::

            Choose between which of QpuManagementProtocol and
            HandleCommBlockForOneNodeProtocol to use and update the first line of
            this docstring and all relevant calls to the protocol in this function.
            When you have made this choice format the name
            correctly with backticks and the full import name, etc.
        """
        if self.ent_successful:
            self.send_signal(self.ent_ready_label)
        elif not self.ent_successful:
            self.send_signal(self.ent_failed_label)

    def handshake(self):
        """
        Implements a handshake protocol to facilitate synchronisation between
        QPUs prior to generating entanglement

        Returns
        -------
        None.

        .. todo::

            Think more on an appropriate wait time.
        """
        while True:
            if self.role == "client":
                # TO DO: think more on an appropriate wait time. Right now I use
                # the 2-qubit gate time for IonQ Aria (rather arbitrarily)
                wait_time = 600 * 10**3  # ns
                self.classical_conn_port.tx_output(self.handshake_ready_label)
                # await response from server or timeout
                evexpr = yield (
                    self.await_port_input(self.classical_conn_port)
                    | self.await_timer(wait_time)
                )
                # note: timeout may occur if physical layers signalled at different
                # times by their InterpreterProtocol
                if evexpr.first_term.value:
                    (msg,) = self.classical_conn_port.rx_input().items
                    if msg == self.handshake_not_ready_label:
                        # resend handshake message
                        self.classical_conn_port.tx_output(self.handshake_ready_label)
                    # checking that the correct message has been received and that
                    # this message is not erroneously sent here from another
                    # protocol:
                    elif msg == self.handshake_ready_label:
                        break  # break the handshake protocols while loop to
                        # allow us to finish handshake
                    else:
                        raise ValueError(
                            "The message must be a str with value "
                            '"ENT_READY_LABEL" or '
                            '"ENT_NOT_READY_LABEL". The received '
                            "message has value {msg}, which does not "
                            "meet the criteria."
                        )
                elif evexpr.second_term.value:
                    # if timeout occurred:
                    # resend handshake message
                    self.classical_conn_port.tx_output(self.handshake_ready_label)
            elif self.role == "server":
                yield self.await_port_input(self.classical_conn_port)
                # TO DO: get signal results here to bring in time?
                if self.ready4ent:
                    self.classical_conn_port.tx_output(self.handshake_ready_label)
                elif not self.ready4ent:
                    self.classical_con_port.tx_output(self.handshake_not_ready_label)
                break  # break the handshake protocols while loop to allow
                # us to finish handshake
        # TO DO: add timed release functionality for synchronisation (which
        # is to be used if timed_release parameter is set to True). The time
        # should relate to the expected latency of the classical message. This
        # could be achieved by having the client node include the time it sends
        # the meeting and have the server infer the
        # channel latency from the time the message
        # is received.
        # TO DO: facilitate having multiple entanglements on the same time slice,
        # which may require having some sort of job ID.


# Following decorator explicitly labels the next class as a subprotocol of
# PhysicalLayerProtocol
@PhysicalLayerProtocol.register
class AbstractCentralSourceEntangleProtocol(PhysicalLayerProtocol):
    """Abstract protocol for generating entanglement.

    An abstract way of generating ebits designed to act on QPUs connected by
    a :class:`~dqc_simulator.hardware.connections.BlackBoxEntanglingQsourceConnection`.
    Details of the ebit generation are ignored, with the state specified
    analytically. The state definition can be used to specify the
    noise on the entangled photons if desired. In this way, the need for
    multiple simulation runs (which would be required by a probabilistic noise
    model) is circumvented.

    Parameters
    ----------
    node : :class:`~netsquid.nodes.node.Node` or None, optional
        Node this protocol runs on. If None, a node should be set later before
        starting this protocol. [1]_
    name : str or None, optional
        Name of protocol. If None, the name of the class is used. [1]_
    role : str
        Whether the QPU is a 'client' (initiating the handshake) or a
        'server' (responding to the handshake.)
        The name of the other QPU node involved in the entangled link. If
        None, must be overwritten prior to the start of the protocol, eg in the
        link layer_protocol.
    other_node_name : str or None, optional
        The name of the other QPU node involved in the entanglment distribution.
        If None, must be overwritten prior to the the start of the protocol,
        eg in the link layer protocol
    comm_qubit_indices = list of int or None, optional
        The index of the memory position in which the comm-qubit that should
        host the entanglement is being held.
    ready4ent = bool por None , optional
        Whether the QPU is ready or not for entanglement. If None, must be
        overwritten prior to the the start of the protocol, eg in the
        link layer protocol.

    Attributes
    ----------
    superprotocol : :class:`netsquid.protocols.nodeprotocols.NodeProtocol`
        The superprotocol for this protocol. This protocol will be
        executed as a subprotocol of the superprotocol. The value of this
        attribute should be overwritten by the superprotocol.
    num_entanglements2generate : int
        The number of instances of the requested entanglement to be specified.
    entanglement_type2generate : str
        The type of entanglement to be used.
    ent_ready_label : str
        The label to use to signal the successful distribution of
        entanglement.
    ent_failed_label : str
        The label to use to signal that entanglement distribution has
        failed

    Notes
    -----
    This abstracts from the details of photon generation by treating flying
    and communication qubits as the same thing. Restraints on the number of
    communication qubits can be enforced at the QPU nodes but entangled
    communication qubits are generated at a central quantum source and sent
    to the QPUs. In this way, we can model error and loss but needn't simulate
    the details of entanglement between static communication qubits and photons.

    References
    ----------
    The parameters in which Ref. [3]_ was cited were inherited from
    :class:`~netsquid.protocols.nodeprotocols.NodeProtocol` and the description
    used for those parameters was taken from the NetSquid documentation [3]_
    """

    def __init__(
        self,
        node=None,
        name=None,
        role=None,
        other_node_name=None,
        comm_qubit_indices=None,
        ready4ent=True,
    ):
        super().__init__(
            node=node,
            name=name,
            role=role,
            other_node_name=other_node_name,
            comm_qubit_indices=comm_qubit_indices,
            ready4ent=ready4ent,
        )
        self.entanglement_type2generate = "bell_pair"
        self.ent_request_msg = "ENT_REQUEST"
        # defining value of read-only deterministic property (see base class)
        self._deterministic = True

    # =============================================================================
    #     def one_way_handshake(self):
    #         """
    #         Implements a one-way version of the handshake protocol.
    #
    #         This subgenerator assumes that entanglement is possible and time
    #         synchronisation is not necessary, which is useful for the
    #         detereministic entanglement distribution scheme here.
    #         """
    #         if self.role == 'client':
    #             self.classical_conn_port.tx_output(self.handshake_ready_label)
    # # =============================================================================
    # #             #await response from server
    # #             yield self.await_port_input(self.classical_conn_port)
    # # =============================================================================
    #         elif self.role == 'server':
    #             yield self.await_port_input(self.classical_conn_port)
    # =============================================================================

    def handle_quantum_input(self):
        """
        Routes qubits on a QPU node to the correct memory position upon
        receiving input to the port this method is bound to.

        Modifies a quantum message inputted to a port, with metadata that puts
        qubits in the correct memory positions and forwards the altered message
        to the qin port of a
        :class:`~netsquid.components.qmemory.QuantumMemory` or
        :class:`~netsquid.components.qprocessor.QuantumProcessor`
        """
        # TO THINK about whether to move this to an abstract base class
        # for protocols expecting a quantum input or the more general
        # physical layer base class (probably the former)
        yield self.await_port_input(self.ent_conn_port)
        msg = self.ent_conn_port.rx_input()  # Message containing qubits
        # adding information on the quantum memory positions to add qubits to
        msg.meta["qm_positions"] = self.comm_qubit_indices
        self.ent_conn_port.forward_input(self.node.qmemory.ports["qin"])
        self.ent_conn_port.tx_input(msg)
        # signalling InterpreterProtocol that entanglement is ready.
        self.send_signal(self.ent_ready_label)
        # undoing the forward input setup above so that, the next time
        # entanglement is distributed, the message is again amended with
        # metadata before forwarding it on
        # =============================================================================
        #         yield self.await_port_input(self.node.qmemory.ports['qin'])
        # =============================================================================
        self.ent_conn_port.forward_input(None)

    def run(self):
        while True:
            # deferring handling of signalling to the base class as this will be
            # identical for all subclasses. This will override the role,
            # other_node_name, comm_qubit_indices, num_entanglements2generate,
            # and entanglement_type2generate attributes with useful values (in
            # place of the default None).
            yield from super().run_link_layer()
            # =============================================================================
            #             yield from self.one_way_handshake()
            # =============================================================================
            yield from self.handshake()
            if self.role == "client" and self.ready4ent:
                # sending entanglement request to quantum source
                self.ent_conn_port.tx_output(self.ent_request_msg)
                # the sending of the request must be done by the client because
                # the client receiving a message back from the server is the
                # last part of the handshake
            yield from self.handle_quantum_input()


class ProbabilisticEntanglingProtocol(PhysicalLayerProtocol):
    """
    An abstract base class for probabilistic physical layer protocols.

    Parameters
    ----------
    node : :class:`~netsquid.nodes.node.Node` or None, optional
        Node this protocol runs on. If None, a node should be set later before
        starting this protocol. [1]_
    name : str or None, optional
        Name of protocol. If None, the name of the class is used.
    role : str or None, optional
        Whether the QPU is a 'client' (initiating the handshake) or a
        'server' (responding to the handshake). If None, must be overwritten
        prior to the start of the protocol, eg in the link layer protocol.
    other_node_name : str or None, optional. [1]_
        The name of the other QPU node involved in the entangled link. If
        None, must be overwritten prior to the start of the protocol, eg in the
        link_layer_protocol.
    comm_qubit_indices = list of int or None, optional
        The index of the memory position in which the comm-qubit that should
        host the entanglement is being held.
    ready4ent = bool, optional
        Whether the QPU is ready or not for entanglement
    max_num_ent_attempts = bool or none, optional
        The number of entanglement attempts to make before giving up and
        handing back control to higher layers.

    Attributes
    ----------
    superprotocol : :class:`netsquid.protocols.nodeprotocols.NodeProtocol`
        The superprotocol for this protocol. This protocol will be
        executed as a subprotocol of the superprotocol. The value of this
        attribute should be overwritten by the superprotocol.
    num_entanglements2generate : int
        The number of instances of the requested entanglement to be specified.
        For many subclasses this will be fixed but in others it will be
        variable and will be overwritten by the link layer.
    entanglement_type2generate : str
        The type of entanglement to be used. For many subclasses this will be
        fixed but in others it will be variable and will be overwritten by the
        link layer.
    ent_ready_label : str
        The label to use to signal the successful distribution of
        entanglement.
    ent_failed_label : str
        The label to use to signal that entanglement distribution has
        failed

    Notes
    -----
    Although num_entanglements2generate and entanglement_type2generate to
    generate will not be used in many subclasses of this protocol, they are
    useful for the user to be able to find.

    References
    ----------
    The parameters in which Ref. [3]_ was cited were inherited from
    :class:`~netsquid.protocols.nodeprotocols.NodeProtocol` and the description
    used for those parameters was taken from the NetSquid documentation. [3]_

    .. todo::

        Think about whether to set role, other_node_name, comm_qubit_indices,
        and read4ent as attributes or properties only, not parameters, as they
        should be set by protocols at higher layers. The advantage of this is
        to avoid confusion when a user sets values for these parameters
        and they are then overwritten. However, the parameter formulation does
        make instantiating the physical layer within higher layers easier.
    """

    def __init__(
        self,
        node=None,
        name=None,
        role=None,
        other_node_name=None,
        comm_qubit_indices=None,
        ready4ent=True,
        max_num_ent_attempts=None,
    ):
        super().__init__(
            node=node,
            name=name,
            role=role,
            other_node_name=other_node_name,
            comm_qubit_indices=comm_qubit_indices,
            ready4ent=ready4ent,
        )
        # defining value of read-only deterministic property (see base class)
        self._deterministic = False
        self.max_num_ent_attempts = max_num_ent_attempts


class MidpointHeraldingProtocol(ProbabilisticEntanglingProtocol):
    """
    Physical layer protocol for generating entanglement between QPUs.

    This protocol assumes that each QPU produces a photon entangled with one
    of its comm qubits and these photons are measured with a BSM between the
    two QPUs to leave the static comm qubits entangled with each other.

    Parameters
    ----------
    node : :class:`~netsquid.nodes.node.Node` or None, optional
        Node this protocol runs on. If None, a node should be set later before
        starting this protocol. [1]_
    name : str or None, optional
        Name of protocol. If None, the name of the class is used. [1]_
    role : str
        Whether the QPU is a 'client' (initiating the handshake) or a
        'server' (responding to the handshake.)
        The name of the other QPU node involved in the entangled link. If
        None, must be overwritten prior to the start of the protocol, eg in the
        link layer_protocol.
    other_node_name : str or None, optional
        The name of the other QPU node involved in the entanglment distribution.
        If None, must be overwritten prior to the the start of the protocol,
        eg in the link layer protocol
    comm_qubit_indices = list of int or None, optional
        The index of the memory position in which the comm-qubit that should
        host the entanglement is being held.
    ready4ent = bool por None , optional
        Whether the QPU is ready or not for entanglement. If None, must be
        overwritten prior to the the start of the protocol, eg in the
        link layer protocol.
    max_num_ent_attempts = bool or none, optional
        The number of entanglement attempts to make before giving up and
        handing back control to higher layers.

    Attributes
    ----------
    superprotocol : :class:`netsquid.protocols.nodeprotocols.NodeProtocol`
        The superprotocol for this protocol. This protocol will be
        executed as a subprotocol of the superprotocol. The value of this
        attribute should be overwritten by the superprotocol.
    num_entanglements2generate : int
        The number of instances of the requested entanglement to be specified.
    entanglement_type2generate : str
        The type of entanglement to be used.
    ent_ready_label : str
        The label to use to signal the successful distribution of
        entanglement.
    ent_failed_label : str
        The label to use to signal that entanglement distribution has
        failed

    Notes
    -----
    Similar idea to the Wehner stack [1]_, [2]_.
    """

    def __init__(
        self,
        node=None,
        name=None,
        role=None,
        other_node_name=None,
        comm_qubit_indices=None,
        ready4ent=True,
        max_num_ent_attempts=100,
    ):
        super().__init__(
            node=node,
            name=name,
            role=role,
            other_node_name=other_node_name,
            comm_qubit_indices=comm_qubit_indices,
            ready4ent=ready4ent,
            max_num_ent_attempts=max_num_ent_attempts,
        )
        self.entanglement_type2generate = "bell_pair"
        self.ent_request_msg = "ENT_REQUEST"
        # The next two instance attributes will be overwritten in the run
        # method because this is the first time that self.node must have a value
        # other than None and it is convenient to use self.node to instantiate
        # the next two attributes.
        self.entangling_connection_port_name = None
        self.ent_conn_port = None
        # initialising internal counter to count the number of times
        # entanglement has been attempted
        self._ent_attempt_counter = 0

    def entangle_comm_qubits(self):
        while True:  # will be broken with break command on successful
            # entanglement or when max number of entanglement attempts
            # is exceeded:
            emit_program = QuantumProgram()
            # By construction, self.comm_qubit_indices are the relevant comm-qubit
            # positional indices to act on rather than the positions of ALL
            # comm-qubits
            emit_program.apply(instr.INSTR_INIT, qubit_indices=self.comm_qubit_indices)
            photon_index = self.node.qmemory.photon_positions[0]
            if len(self.comm_qubit_indices) == 1:
                comm_qubit_index = self.comm_qubit_indices[0]
            else:
                raise ValueError(
                    "There are too many comm-qubit indices. "
                    "Currently, only one comm-qubit can be used here."
                    "It is assumed that there is one photon emission"
                    " at a time."
                )
            emit_program.apply(
                instr.INSTR_EMIT, qubit_indices=[comm_qubit_index, photon_index]
            )
            # =============================================================================
            #         #initialising a COPY of comm_qubit indices
            #         unused_comm_qubit_indices = list(self.comm_qubit_indices)
            #         #TO DO: decide if the following should be an attribute of the qmemory
            #         #or the node and add the attributed to whichever you decide on
            #         photon_indices = self.node.qmemory.photon_positions
            #         for ii in range(self.num_entanglements2generate):
            #             comm_qubit_index = unused_comm_qubit_indices[ii]
            #             photon_index = photon_indices[ii]
            #             emit_program.apply(instr.INSTR_EMIT,
            #                           qubit_indices=[comm_qubit_index, photon_index])
            # =============================================================================
            yield self.node.qmemory.execute_program(emit_program)
            # wait on BSM outcome
            yield self.await_port_input(self.ent_conn_port)
            (bsm_outcome,) = self.ent_conn_port.rx_input().items
            program = QuantumProgram()
            if bsm_outcome.success:
                # converting to PHI_PLUS state
                # applying corrective gates to only one of the qubits involved in
                # the bell pair (which one does not matter up to a global phase)
                if self.role == "server":
                    # x gate needed for both PSI_PLUS and PSI_MINUS
                    program.apply(instr.INSTR_X, qubit_indices=comm_qubit_index)
                    if bsm_outcome.bell_index == ks.BellIndex.PSI_MINUS:
                        program.apply(instr.INSTR_Z, qubit_indices=comm_qubit_index)
                    elif bsm_outcome.bell_index != ks.BellIndex.PSI_PLUS:
                        raise ValueError(
                            "the BSMOutcome must have bell_index attr"
                            "with value netsquid.qubits.ketstates."
                            "BellIndex.PSI_PLUS or PSI_MINUS. Instead, "
                            f"it has value {bsm_outcome.bell_index} here."
                        )
                    yield self.node.qmemory.execute_program(program)
                    self.classical_conn_port.tx_output(self.ent_ready_label)
                    self.send_signal(self.ent_ready_label)
                    break
                elif self.role == "client":
                    yield self.await_port_input(self.ent_conn_port)
                    self.send_signal(self.ent_ready_label)
                    break  # breaking the while loop at the start of this method
            else:  # if bsm_outcome.success == False:
                if self._ent_attempt_counter > (self.max_num_ent_attempts - 1):
                    self.send_signal(self.ent_failed_label)
                    break  # breaking the while loop at the start of this method
                else:
                    self._ent_attempt_counter = self._ent_attempt_counter + 1

    def run(self):
        while True:
            # deferring handling of signalling to the base class as this will be
            # identical for all subclasses. This will override the role,
            # other_node_name, comm_qubit_indices, num_entanglements2generate,
            # and entanglement_type2generate attributes with useful values (in
            # place of the default None).
            yield from super().run_link_layer()
            yield from self.handshake()  # TO DO: improve handshake
            # At the moment, the following will only be able to emit one photon
            # at a time. INSTR_EMIT can only emit one photon at a time but one
            # could use many such instructions in a program to emit more than
            # one.
            yield from self.entangle_comm_qubits()



# -*- coding: utf-8 -*-
"""
Noise models for acting on simulated QPUs.

.. todo::

    Deprecate the analytical models. The analytical models arose from a
    misunderstanding about what NetSquid's default DepolarNoiseModel was
    doing (which is already analytical in the DM formalism.) As such, these
    models do the same thing as NetSquid's built in model but much slower
    and less flexibly. They are retained with a DeprecationWarning for now
    to retain backwards compatibility.
"""

import functools as ft
import itertools as it
import warnings

import netsquid as ns
from netsquid.components.models.qerrormodels import QuantumErrorModel
from netsquid.qubits import qubitapi as qapi, operators as ops
from netsquid.util.constrainedmap import ValueConstraint
from netsquid.qubits.dmutil import partialtrace, reorder_dm
from netsquid.qubits.qformalism import QFormalism
import numpy as np

# To take analytical approach of applying a channel rather than an operator,
# you will need to specify that the density matrix formalism is used
# and also work within that formalism specifically. This means you need
# to access the dm itself.


def apply_analytical_depolarisation2dm(qubits, p_error):
    """
    Helper function

    .. deprecated::
        Will be deprecated in future version as `netsquid.qubitapi.depolarise`
        already does the same job
        when the QFormalism is set to DM! This is kept only for backwards
        compatibility.

    Assumes that :class:`~netsquid.qubits.dmtools.DenseDMRepr` is used.

    Parameters
    ----------
    qubits : :class:`~netsquid.qubits.qubit.Qubit` obj or list of them
        The qubits upon which to act depolarising noise.
    p_error : float
        The probability of a qubit being depolarised in this channel. See
        the notes section.

    Notes
    -----
    Applies a depolarisation channel of the form:
    .. :math:
            `\rho_out = (1 - p_error) rho_{in}
           + p_error/dim{i, j} Tr_{i,j} (rho_in) o_times I_{i, j}'.

    This circumvents the need for repeated runs, as it is not probabilistic.
    It is equivalent to equations (2) and (3) in
    QuantumRepeaters:The Role of Imperfect Local Operations in Quantum
    Communication
    by Briegel, Dur, Cirac and Zoller but I split it into two parts.
    I assume the ideal operation is first done and then act a channel which
    either does nothing with some probability or decoheres the ideal output
    with some probability.

    This noise model is designed to be used after a gate or only on one qubit
    at a time because it assumes all qubits share a qstate. This assumption
    is good if the qubits have been acted on by a multi-qubit gate
    """
    warnings.warn(
        "This will be deprecated in a future version. Use NetSquid's"
        " built-in netsquid.qubitapi.depolarise function instead.",
        DeprecationWarning,
    )
    if isinstance(qubits, ns.qubits.qubit.Qubit):
        # =============================================================================
        #         print(f'The qubit {qubits} has state {qubits.qstate.qrepr}')
        # =============================================================================
        qubits = [qubits]
    elif len(qubits) > 1:
        # =============================================================================
        #         print(f'The qubits {qubits} have states {qubits[0].qstate.qrepr}'
        #               f'{qubits[1].qstate.qrepr}')
        # =============================================================================
        ii = 0
        while ii < (len(qubits) - 1):
            qubit = qubits[ii]
            nxt_qubit = qubits[ii + 1]
            if qubit.qstate != nxt_qubit.qstate:
                raise Exception(
                    f"Qubits {qubit.name} and {nxt_qubit.name} do"
                    f" not share a quantum state"
                )
            else:
                ii = ii + 1
    else:
        raise TypeError(f"qubits must be type {ns.qubits.qubit.Qubit} or list")
    qubit = qubits[0]
    indices4entire_state = qubit.qstate.indices  # dictionary
    indices4subspace = []  # subspace containing qubits involved in ideal gate
    for qubit in qubits:
        index4qubit = indices4entire_state[qubit.name]
        indices4subspace = indices4subspace + [index4qubit]
    ideal_dm = qubit.qstate.dm  # does not matter which qubit as all share same
    # qstate after gate which acts on all of them
    subspace_dim = 2 ** len(indices4subspace)
    noise_mat = np.kron(partialtrace(ideal_dm, indices4subspace), np.eye(subspace_dim))
    # =============================================================================
    #     print(f'The noise mat is {noise_mat}')
    # =============================================================================
    # re-ordering so that the noise matrix applies noise to the correct places:
    # list of form [complement_indices, subspace_indices] needed for noise_mat
    # to apply to the right places
    indices4entire_state_list = list(indices4entire_state.values())
    indices4complement_space = list(
        set(indices4entire_state_list) - set(indices4subspace)
    )
    desired_ordering = [*indices4complement_space, *indices4subspace]
    noise_mat = reorder_dm(noise_mat, desired_ordering)
    partially_depolarised_dm = (1 - p_error) * ideal_dm + (
        p_error / subspace_dim
    ) * noise_mat
    # =============================================================================
    #     print(f'after depol {partially_depolarised_dm} with size {np.shape(partially_depolarised_dm)}')
    # =============================================================================
    qubit_list = qubit.qstate.qubits
    qapi.assign_qstate(qubit_list, partially_depolarised_dm, formalism=QFormalism.DM)
    # TO DO: add error if dm is not a valid dm
    # TO DO: Neaten and optimise this function


# The following class is based on
# netsquid.components.models.qerrormodels.DepolarNoiseModel and uses some code
# from there in accordance with the license agreement which allows such actions
# for non-commercial research use. It follows the standard structure of a
# QuantumErrorModel
class AnalyticalDepolarisationModel(QuantumErrorModel):
    """
    Noise model for applying analytical depolarising noise to a qubit.

    .. deprecated::
        Will be deprecated in future version as
        `netsquid.components.models.qerrormodels.DepolarNoiseModel` already
        does the same job when the DM formalism is used.

    Based on :class:`~netsquid.components.models.DepolarNoiseModel` but
    adjusted to
    be applied analytically and within the DM formalism only. This is
    intended to avoid the need for multiple runs and thus reduce
    computation time. However, this model is inappropriate for systems with
    many qubits because of the restriction to the memory-intensive DM
    formalism.

    Parameters
    ----------
    p_error : float
        Probability that the qubits are depolarised. If `time_independent`
        is True (default), then this a probability as defined in
        .. :math:
                `\rho_out = (1 - p_error) rho_in
               + p_error/dim{i, j} Tr_{i,j} (rho_in) o_times I_{i, j}'.

        If `time_independent` is False,
        then p_{error} is the exponential depolarizing rate per unit time
        [Hz], such that
        .. :math:
                `\rho_{out} = (1 - p_{error}) rho_{in}
               + prob_{error}/dim{i, j} Tr_{i,j} (rho_{in})
               o_times I_{i, j}'

        with :math: `prob_{error} = 1 - e^{\delta t p_{error} * e^{-9}}',
        where :math: `\delta t' is the time qubits spend on a component
        (eg, a quantum memory).
    time_independent : bool, optional
        Whether or not the probability of a depolarisation error occurring
        depends on time. The default is True.
    """

    def __init__(self, p_error, time_independent=True, **kwargs):
        warnings.warn(
            "This will be deprecated in a future version. Use NetSquid's"
            " built-in "
            "netsquid.components.models.qerrormodels.DepolarNoiseModel"
            " instead.",
            DeprecationWarning,
        )
        super().__init__(**kwargs)
        # NOTE time independence should be set *before* the rate
        self.add_property("time_independent", time_independent, value_type=bool)

        def prob_constraint(value):
            if self.time_independent and not 0 <= value <= 1:
                return False
            elif value < 0:
                return False
            else:
                return True

        self.add_property(
            "p_error",
            value=p_error,
            value_type=(int, float),
            value_constraints=ValueConstraint(prob_constraint),
        )

    @property
    def p_error(self):
        return self.properties["p_error"]

    @p_error.setter
    def p_error(self, value):
        self.properties["p_error"] = value

    @property
    def time_independent(self):
        """bool: Whether the probability of depolarising is time independent."""
        return self.properties["time_independent"]

    @time_independent.setter
    def time_independent(self, value):
        self.properties["time_independent"] = value

    def error_operation(self, qubits, delta_time=0, **kwargs):
        """
        Error to apply to qubits

        INPUT:
            qubits : tuple of :obj:`~netsquid.qubits.qubit.Qubit`
                Qubits to apply noise to.
            delta_time : float optional
                Time qubits have spent on a component in ns (ns used to
                maintain consistency with QuantumErrorModel parent class).

        """
        if self.time_independent:  # for adding errors to quantum
            # gates or other events which can be thought
            # of as discrete. Thus, this models things
            # like imperfectly applied gates.
            # I assume here that the qubits all share a
            # qstate attribute
            # (netsquid.qubits.Qubit.qstate), which is the
            # case after they are acted on by a shared
            # gate. p_error is a probability here
            apply_analytical_depolarisation2dm(qubits, self.p_error)
        else:  # for memory decoherence. p_error is now a rate in Hz
            for qubit in qubits:
                prob_error = 1.0 - np.exp(-delta_time * 1e-9 * self.p_error)
                # factor of 1e-9 in exponent is to convert delta_time from ns to
                # s, so that the units cancel with p_error (which is in Hz)
                apply_analytical_depolarisation2dm(qubit, prob_error)


# =============================================================================
#                 # for DEBUGGING only. REMOVE once finished:
#                 print(f"qubit {qubit} has state{ns.qubits.reduced_dm(qubit)}")
#
# =============================================================================


class MeasurementNoiseModel(QuantumErrorModel):
    """
    Noise model for applying bit flip noise to a measurement, giving the
    opposite result from the ideal measurement with some probability.
    """

    def __init__(self, error_probability, **kwargs):
        super().__init__(**kwargs)

        def error_probability_constraint(value):
            if not 0 <= value <= 1:
                return False
            else:
                return True

        self.add_property(
            "error_probability",
            error_probability,
            value_type=(int, float),
            value_constraints=ValueConstraint(error_probability_constraint),
        )

    @property
    def error_probability(self):
        """float:
        probability that a bit flip error will occur during measurement."""
        return self.properties["error_probability"]

    @error_probability.setter
    def error_probability(self, value):
        self.properties["error_probability"] = value

    def error_operation(self, qubits, delta_time=0, **kwargs):
        """Error operation to apply to qubits.

        Parameters
        ----------
        qubits : tuple of :obj:`~netsquid.qubits.qubit.Qubit`
            Qubits to apply noise to.
        """
        for qubit in qubits:
            if qubit is not None:
                qapi.apply_pauli_noise(
                    qubit, (1 - self.error_probability, self.error_probability, 0, 0)
                )


class NDimDepolarNoiseModel(QuantumErrorModel):
    """
    Apply N dimensional depolar noise channel.

    Useful for example for depolarisation noise due to multi-qubit gates.
    """

    def __init__(self, error_probability, **kwargs):
        super().__init__(**kwargs)

        def error_probability_constraint(value):
            if not 0 <= value <= 1:
                return False
            else:
                return True

        self.add_property(
            "error_probability",
            error_probability,
            value_type=(int, float),
            value_constraints=ValueConstraint(error_probability_constraint),
        )

    @property
    def error_probability(self):
        """float:
        probability that a bit flip error will occur during measurement."""
        return self.properties["error_probability"]

    @error_probability.setter
    def error_probability(self, value):
        self.properties["error_probability"] = value

    def error_operation(self, qubits, delta_time=0, **kwargs):
        """Error operation to apply to qubits.

        Parameters
        ----------
        qubits : tuple of :obj:`~netsquid.qubits.qubit.Qubit`
            Qubits to apply noise to.
        """
        # Defining the number of possible Pauli strings that could be applied
        # to the qubits, including the identity string
        num_paulis = 4 ** len(qubits)
        num_errors = num_paulis - 1

        # Defining all Pauli strings that can act on the relevant qubits
        paulis = [
            ft.reduce(lambda a, b: a ^ b, tup)
            for tup in it.product([ops.I, ops.X, ops.Y, ops.Z], repeat=len(qubits))
        ]

        # Apply Pauli strings to relevant qubits
        p = self.error_probability / num_paulis
        qapi.stochastic_operate(
            qubits, paulis, p_weights=tuple([1 - num_errors * p] + [p] * num_errors)
        )



"""
Tools to convert a monolithic quantum circuit specified in openQASM 2.0 to
gate tuples representing a partitioned quantum circuit.

References
----------
.. [1] A Cross, et al, Open Quantum Assembly Language, arXiv:1707.03429.
"""

# from dqc_simulator.software.ast2dqc_circuit import Ast2DqcCircuitTranslator
from dqc_simulator.software.compiler_preprocessing import (
    preprocess_qasm_to_compilable_monolithic as preprocess,
)
from dqc_simulator.software.partitioner import (
    first_come_first_served_qubits_to_qpus as allocate,
    partition_gate_tuples as partition,
)


def qasm2gate_tuples(dqc, filepath, scheme, include_path="."):
    """
    Converts a monolithic quantum circuit specified in openQASM 2.0 to
    gate tuples representing a partitioned quantum circuit.

    dqc : :class:`dqc_simulator.hardware.dqc_creation.DQC`
        A DQC network.
    filepath : str
        The path to the .qasm file.
    scheme : str
        The type of remote gate to use. Allowed values are 'cat', '1tp',
        '2tp', and 'tp_safe'.
    include_path : str
        The path to the directory containing any include (.inc) files imported
        in the .qasm file. See [1]_. The default is '.', which looks in the
        current working directory.
    """
    # import .qasm file and convert to gate_tuples for monolithic_circuit
    dqc_circuit = preprocess(filepath, include_path=include_path)
    monolithic_circuit = dqc_circuit.ops  # gate_tuples

    # Determine allocation of processing qubits to QPUs
    old_to_new_lookup, proc_qubit_allocation4each_qpu = allocate(
        monolithic_circuit, list(dqc.nodes.values())
    )

    # Partition according to the previously defined qubit allocation
    partitioned_gate_tuples = partition(
        monolithic_circuit,
        dqc,  # defined earlier in tutorial
        scheme,
        old_to_new_lookup,
        proc_qubit_allocation4each_qpu,
    )
    return partitioned_gate_tuples



# -*- coding: utf-8 -*-
# =============================================================================
# Created on Wed Sep 20 14:28:30 2023
#
# @author: kenny
# =============================================================================
"""
Provides helper functions.

The helper functions are intended to smooth the process of working with the
simulator and also aid development.
"""

# This module provides utility functions for smoothing
# the process of working with the simulator.

from enum import Enum
import inspect

import pydynaa
from netsquid.qubits.dmtools import DenseDMRepr
from netsquid.protocols.protocol import Signals
from netsquid.qubits import qubitapi as qapi
from netsquid.util.datacollector import DataCollector


def create_wrapper_with_some_args_fixed(func, args2fix, **kwargs2fix):
    """Bakes in some arguments to a function,

    Parameters
    ----------
    func : function
        The function to create a wrapper for.
    args2fix : dict or None
        The relative postion of the argument (key in dict) that should be fixed
        and the value to fix it to (value in dict).
    kwargs2fix
        The keyword arguments to fix (specified in the normal way not
        aggregated)

    Returns
    -------
    wrapper : function
        A wrapper for `func` which fixes some of the arguments, so that the
        user inputs a reduced set of arguments relative to func.
    """

    def func_wrapper(*unfixed_args, **kwargs):
        if args2fix is None:
            wrapper = func(*unfixed_args, **kwargs2fix, **kwargs)
        else:
            unfixed_arg_iter = iter(unfixed_args)
            num_pos_args4func = len(args2fix) + len([*unfixed_args])
            wrapper = func(
                *(
                    args2fix[ii] if ii in args2fix else next(unfixed_arg_iter)
                    for ii in range(num_pos_args4func)
                ),
                **kwargs2fix,
                **kwargs,
            )
        return wrapper

    return func_wrapper


def filter_kwargs4internal_functions(funcs_with_kwargs, kwargs):
    """
    Filters keyword arguments specified by a parent function for use in calls
    to functions within the definition of the parent function.

    Parameters
    ----------
    funcs_with_kwargs : list or tuple of functions
        The internal functions needing keyword arguments
    kwargs : dict
        The keyword arguments.

    Returns
    -------
    sorted_kwargs : dict of dicts
        The kwargs for each different function. Key is function and values are
        kwargs to be unpacked in each function call.
    """
    sorted_kwargs = {}
    for func in funcs_with_kwargs:
        # making list of positional and keyword arguments
        arg_names = list(inspect.signature(func).parameters)
        # filtering for only the kwargs and associating them with their values
        kwargs4func = {
            arg_name: kwargs[arg_name] for arg_name in arg_names if arg_name in kwargs
        }
        sorted_kwargs[func] = kwargs4func
    return sorted_kwargs


# =============================================================================
# #_idx for debugging only REMOVE ONCE debugged
# from netsquid.qubits.qubit import Qubit
# def _idx(qubits):
#     # This helper function is always executed together with _qrepr, so we skip safety checks
#     # and combining.
#     if isinstance(qubits, Qubit):
#         return qubits.qstate.indices_of([qubits])
#     elif isinstance(qubits, list):
#         return qubits[0].qstate.indices_of(qubits)
#     else:
#         raise TypeError("The qubits must be given as a list or a single Qubit.")
# =============================================================================


def get_data_collector(master_protocol, qubit_indices_2b_checked, desired_state):
    """Sets up data collector for arbitrary simulation.

    Parameters
    ----------
    master_protocol: :class:netsquid.protocols.protocol.Protocol
        Protocol to run the DQC circuit.
    qubits_2b_checked: list of tuples of form (list of indices, node)
        The qubits whose state should be checked against a known state
    desired_state: :class:numpy.ndarray
        The ideal state which the actual state should be compared to. Can be
        given as a ket vector or density matrix regardless of the formalism
        being worked with

    Returns
    -------
    dc : :class:`~netsquid.util.datacollector.DataCollector`
        Data collector to record fidelity.
    """

    def collect_fidelity_data(evexpr):
        # defining which protocol you wish to sample the event expression from
        # (CorrectionProtocol)
        qubits_2b_checked = []
        for qubit_node_tuple in qubit_indices_2b_checked:
            if isinstance(qubit_node_tuple[0], int):
                qubit_indices = [qubit_node_tuple[0]]
            elif isinstance(qubit_node_tuple[0], list):
                qubit_indices = qubit_node_tuple[0]
            else:
                raise TypeError(
                    f"{qubit_node_tuple} not of correct form."
                    "The first element of each tuple must be qubit"
                    " index or list of qubit indices"
                )
            node = qubit_node_tuple[-1]
            qubits = node.qmemory.pop(qubit_indices)
            qubits_2b_checked = qubits_2b_checked + qubits

        fidelity = qapi.fidelity(qubits_2b_checked, desired_state, squared=True)

        for qubit in qubits_2b_checked:
            qapi.discard(qubit)
        return {"fidelity": fidelity}

    dc = DataCollector(collect_fidelity_data)
    dc.collect_on(
        pydynaa.EventExpression(
            source=master_protocol, event_type=Signals.FINISHED.value
        )
    )
    return dc


def get_data_collector4dm(master_protocol, qubit_indices_2b_checked, desired_state):
    """Creates a data collector with a specific configuration.

    Sets up :class:`~netsquid.util.datacollector.DataCollector` for simulation
    in which `qubits_indices_2b_checked` share a dm. This avoids the
    sensitivity to the ordering of the qubits to
    be checked in the list produced below but at the cost of requiring desired
    state to be a density matrix.

    Parameters
    ----------
    master_protocol: :class:`~netsquid.protocols.protocol.Protocol`
        Protocol to run the DQC circuit.
    qubits_2b_checked: list of tuples of form (list of indices, node)
        The qubits whose state should be checked against a known state
    desired_state: array
        The ideal state which the actual state should be compared to. Must be
        given as a density matrix.

    Returns
    -------
    :class:`~netsquid.util.datacollector.DataCollector`
        Data collector to record fidelity.
    """

    def collect_fidelity_data(evexpr):
        # defining which protocol you wish to sample the event expression from
        # (CorrectionProtocol)
        qubits_2b_checked = []
        for qubit_node_tuple in qubit_indices_2b_checked:
            if isinstance(qubit_node_tuple[0], int):
                qubit_indices = [qubit_node_tuple[0]]
            elif isinstance(qubit_node_tuple[0], list):
                qubit_indices = qubit_node_tuple[0]
            else:
                raise TypeError(
                    f"{qubit_node_tuple} not of correct form."
                    "The first element of each tuple must be qubit"
                    " index or list of qubit indices"
                )
            node = qubit_node_tuple[-1]
            qubits = node.qmemory.pop(qubit_indices)
            qubits_2b_checked = qubits_2b_checked + qubits
        qapi.combine_qubits(qubits_2b_checked)
        fidelity = qubits[0].qstate.qrepr.fidelity(
            DenseDMRepr(dm=desired_state), squared=True
        )  # for debugging
        for qubit in qubits_2b_checked:
            qapi.discard(qubit)
        return {"fidelity": fidelity}

    dc = DataCollector(collect_fidelity_data)
    dc.collect_on(
        pydynaa.EventExpression(
            source=master_protocol, event_type=Signals.FINISHED.value
        )
    )
    return dc


class QDCSignals(Enum):
    """
    Specialist signals for use in a quantum data centre (QDC).
    """

    RESULT_PRODUCED = pydynaa.EventType(
        "result_produced", "result of instruction produced and ready to be logged."
    )


def get_data_collector_for_mid_sim_instr_output():
    """
    For collecting instruction output generated during the sim.

    This collects results generated by the `result_produced_evexpr` inside
    the _logged_instr method, which indicates that an instruction has produced
    a result that should be logged. This is most likely a measurement result
    from a measurement instruction.

    Returns
    -------
    :class:`~netsquid.util.datacollector.DataCollector`
        Data collector to record the output of an instruction (most likely a
        measurement result from a measurement instruction).
    """

    def collect_instruction_output(evexpr):
        protocol = evexpr.triggered_events[-1].source
        result_and_index = protocol.get_signal_result(QDCSignals.RESULT_PRODUCED)
        result = result_and_index[0]
        ancilla_qubit_index = result_and_index[1]
        return {"result": result, "ancilla_qubit_index": ancilla_qubit_index}

    dc = DataCollector(collect_instruction_output)
    dc.collect_on(
        [pydynaa.EventExpression(event_type=QDCSignals.RESULT_PRODUCED.value)]
    )
    return dc
